@@ -90,6 +90,8 @@ static int bts_signal_cbfn(unsigned int subsys, unsigned int signal,
 static const struct rate_ctr_desc bts_ctr_desc[] = {
 	[BTS_CTR_PAGING_RCVD] =		{"paging:rcvd", "Received paging requests (Abis)"},
 	[BTS_CTR_PAGING_DROP] =		{"paging:drop", "Dropped paging requests (Abis)"},
+	[BTS_CTR_PAGING_DROP_PS] =	{"paging:drop-ps", "Dropped paging requests (PS/PCU)"},
+	[BTS_CTR_PAGING_CONG] =		{"paging:cong", "Paging congestion detected (Abis)"},
 	[BTS_CTR_PAGING_SENT] =		{"paging:sent", "Sent paging requests (Um)"},
 
 	[BTS_CTR_RACH_RCVD] =		{"rach:rcvd", "Received RACH requests (Um)"},
@@ -126,7 +128,7 @@ static const struct rate_ctr_group_desc cbch_ctrg_desc = {
 	cbch_ctr_desc
 };
 
-static struct osmo_tdef bts_T_defs[] = {
+struct osmo_tdef bts_T_defs[] = {
 	/* T-1: FIXME: Ideally should be dynamically calculated per trx at
 	 * shutdown start based on params below, and highest trx value taken:
 	 * + VTY's power-ramp step-interval.
@@ -141,6 +143,11 @@ static struct osmo_tdef bts_T_defs[] = {
 	{}
 };
 
+struct osmo_tdef abis_T_defs[] = {
+	{ .T=-15, .default_val=0, .unit=OSMO_TDEF_MS, .desc="Time to wait between Channel Activation and dispatching a cached early Immediate Assignment" },
+	{}
+};
+
 static const uint8_t bts_nse_timer_default[] = { 3, 3, 3, 3, 30, 3, 10 };
 static const uint8_t bts_cell_timer_default[] =
 				{ 3, 3, 3, 3, 3, 10, 3, 10, 3, 10, 3 };
@@ -232,6 +239,7 @@ struct gsm_bts *gsm_bts_alloc(void *ctx, uint8_t bts_num)
 
 	bts->T_defs = bts_T_defs;
 	osmo_tdefs_reset(bts->T_defs);
+	osmo_tdefs_reset(abis_T_defs);
 	bts->shutdown_fi = osmo_fsm_inst_alloc(&bts_shutdown_fsm, bts, bts,
 					       LOGL_INFO, NULL);
 	osmo_fsm_inst_update_id_f(bts->shutdown_fi, "bts%d", bts->nr);
@@ -329,14 +337,14 @@ int bts_init(struct gsm_bts *bts)
 	bts->rtp_port_range_end = 17407;
 	bts->rtp_port_range_next = bts->rtp_port_range_start;
 	bts->rtp_ip_dscp = -1;
+	bts->rtp_priority = -1;
 
 	/* Default (fall-back) MS/BS Power control parameters */
-	bts->bs_dpc_params = power_ctrl_params_def;
-	bts->ms_dpc_params = power_ctrl_params_def;
+	power_ctrl_params_def_reset(&bts->bs_dpc_params, true);
+	power_ctrl_params_def_reset(&bts->ms_dpc_params, false);
 
 	/* configurable via OML */
 	bts->load.ccch.load_ind_period = 112;
-	load_timer_start(bts);
 	bts->rtp_jitter_buf_ms = 100;
 	bts->max_ta = 63;
 	bts->ny1 = 4;
@@ -370,6 +378,8 @@ int bts_init(struct gsm_bts *bts)
 	/* features implemented in 'common', available for all models */
 	osmo_bts_set_feature(bts->features, BTS_FEAT_ETWS_PN);
 	osmo_bts_set_feature(bts->features, BTS_FEAT_IPV6_NSVC);
+	osmo_bts_set_feature(bts->features, BTS_FEAT_ABIS_OSMO_PCU);
+	osmo_bts_set_feature(bts->features, BTS_FEAT_DYN_TS_SDCCH8);
 
 	rc = bts_model_init(bts);
 	if (rc < 0) {
@@ -378,7 +388,7 @@ int bts_init(struct gsm_bts *bts)
 	}
 
 	/* TRX0 was allocated early during gsm_bts_alloc, not later through VTY */
-	bts_trx_init(bts->c0);
+	bts_model_trx_init(bts->c0);
 	bts_gsmnet.num_bts++;
 
 	if (!initialized) {
@@ -396,7 +406,7 @@ int bts_init(struct gsm_bts *bts)
 	bts->smscb_queue_tgt_len = 2;
 	bts->smscb_queue_hyst = 2;
 
-	INIT_LLIST_HEAD(&bts->oml_queue);
+	INIT_LLIST_HEAD(&bts->bsc_oml_hosts);
 
 	/* register DTX DL FSM */
 	rc = osmo_fsm_register(&dtx_dl_amr_fsm);
@@ -444,66 +454,6 @@ int bts_link_estab(struct gsm_bts *bts)
 	return bts_model_oml_estab(bts);
 }
 
-/* prepare the per-SAPI T200 arrays for a given lchan */
-static int t200_by_lchan(int *t200_ms_dcch, int *t200_ms_acch, struct gsm_lchan *lchan)
-{
-	struct gsm_bts *bts = lchan->ts->trx->bts;
-
-	/* we have to compensate for the "RTS advance" due to the asynchronous interface between
-	 * the BTS (LAPDm) and the PHY/L1 (OsmoTRX or DSP in case of osmo-bts-{sysmo,lc15,oc2g,octphy} */
-	int32_t fn_advance = bts_get_avg_fn_advance(bts);
-	int32_t fn_advance_us = fn_advance * 4615;
-	int fn_advance_ms = fn_advance_us / 1000;
-
-	t200_ms_acch[DL_SAPI0] = bts->t200_ms[T200_SACCH_SDCCH] + fn_advance_ms;
-	t200_ms_acch[DL_SAPI3] = bts->t200_ms[T200_SACCH_SDCCH] + fn_advance_ms;
-
-	if (lchan->repeated_acch_capability.dl_facch_all && (lchan->type == GSM_LCHAN_TCH_F || lchan->type == GSM_LCHAN_TCH_H)) {
-		t200_ms_acch[DL_SAPI0] *= 2;
-		t200_ms_acch[DL_SAPI3] *= 2;
-	}
-
-	switch (lchan->type) {
-	case GSM_LCHAN_SDCCH:
-		t200_ms_dcch[DL_SAPI0] = bts->t200_ms[T200_SDCCH] + fn_advance_ms;
-		t200_ms_dcch[DL_SAPI3] = bts->t200_ms[T200_SDCCH_SAPI3] + fn_advance_ms;
-		break;
-	case GSM_LCHAN_TCH_F:
-		t200_ms_dcch[DL_SAPI0] = bts->t200_ms[T200_FACCH_F] + fn_advance_ms;
-		t200_ms_dcch[DL_SAPI3] = bts->t200_ms[T200_FACCH_F] + fn_advance_ms;
-		break;
-	case GSM_LCHAN_TCH_H:
-		t200_ms_dcch[DL_SAPI0] = bts->t200_ms[T200_FACCH_H] + fn_advance_ms;
-		t200_ms_dcch[DL_SAPI3] = bts->t200_ms[T200_FACCH_H] + fn_advance_ms;
-		break;
-	default:
-		/* Channels such as CCCH don't use lapdm DL, and hence no T200 is needed */
-		return -1;
-	}
-	return 0;
-}
-
-int lchan_init_lapdm(struct gsm_lchan *lchan)
-{
-	struct lapdm_channel *lc = &lchan->lapdm_ch;
-	int t200_ms_dcch[_NR_DL_SAPI], t200_ms_acch[_NR_DL_SAPI];
-
-	if (t200_by_lchan(t200_ms_dcch, t200_ms_acch, lchan) == 0) {
-		LOGPLCHAN(lchan, DLLAPD, LOGL_DEBUG,
-			  "Setting T200 D0=%u, D3=%u, S0=%u, S3=%u (all in ms)\n",
-			  t200_ms_dcch[DL_SAPI0], t200_ms_dcch[DL_SAPI3],
-			  t200_ms_acch[DL_SAPI0], t200_ms_acch[DL_SAPI3]);
-		lapdm_channel_init3(lc, LAPDM_MODE_BTS, t200_ms_dcch, t200_ms_acch, lchan->type,
-				    gsm_lchan_name(lchan));
-		lapdm_channel_set_flags(lc, LAPDM_ENT_F_POLLING_ONLY);
-		lapdm_channel_set_l1(lc, NULL, lchan);
-	}
-	/* We still need to set Rx callback to receive RACH requests: */
-	lapdm_channel_set_l3(lc, lapdm_rll_tx_cb, lchan);
-
-	return 0;
-}
-
 #define CCCH_RACH_RATIO_COMBINED256      (256*1/9)
 #define CCCH_RACH_RATIO_SEPARATE256      (256*10/55)
 
@@ -853,28 +803,39 @@ struct gsm_time *get_time(struct gsm_bts *bts)
 	return &bts->gsm_time;
 }
 
-int bts_supports_cm(struct gsm_bts *bts, enum gsm_phys_chan_config pchan,
-		    enum gsm48_chan_mode cm)
+int bts_supports_cm(const struct gsm_bts *bts,
+		    const struct rsl_ie_chan_mode *cm)
 {
 	enum osmo_bts_features feature = _NUM_BTS_FEAT;
 
-	/* We assume that signalling support is mandatory,
-	 * there is no BTS_FEAT_* definition to check that. */
-	if (cm == GSM48_CMODE_SIGN)
+	switch (cm->spd_ind) {
+	case RSL_CMOD_SPD_SIGN:
+		/* We assume that signalling support is mandatory,
+		 * there is no BTS_FEAT_* definition to check that. */
 		return 1;
+	case RSL_CMOD_SPD_SPEECH:
+		break;
+	case RSL_CMOD_SPD_DATA:
+	default:
+		return 0;
+	}
 
 	/* Before the requested pchan/cm combination can be checked, we need to
 	 * convert it to a feature identifier we can check */
-	switch (pchan) {
-	case GSM_PCHAN_TCH_F:
-		switch(cm) {
-		case GSM48_CMODE_SPEECH_V1:
+	switch (cm->chan_rt) {
+	case RSL_CMOD_CRT_OSMO_TCH_VAMOS_Bm:
+		if (!osmo_bts_has_feature(bts->features, BTS_FEAT_VAMOS))
+			return 0;
+		/* fall-through */
+	case RSL_CMOD_CRT_TCH_Bm:
+		switch (cm->chan_rate) {
+		case RSL_CMOD_SP_GSM1:
 			feature	= BTS_FEAT_SPEECH_F_V1;
 			break;
-		case GSM48_CMODE_SPEECH_EFR:
+		case RSL_CMOD_SP_GSM2:
 			feature	= BTS_FEAT_SPEECH_F_EFR;
 			break;
-		case GSM48_CMODE_SPEECH_AMR:
+		case RSL_CMOD_SP_GSM3:
 			feature = BTS_FEAT_SPEECH_F_AMR;
 			break;
 		default:
@@ -883,12 +844,16 @@ int bts_supports_cm(struct gsm_bts *bts, enum gsm_phys_chan_config pchan,
 		}
 		break;
 
-	case GSM_PCHAN_TCH_H:
-		switch(cm) {
-		case GSM48_CMODE_SPEECH_V1:
+	case RSL_CMOD_CRT_OSMO_TCH_VAMOS_Lm:
+		if (!osmo_bts_has_feature(bts->features, BTS_FEAT_VAMOS))
+			return 0;
+		/* fall-through */
+	case RSL_CMOD_CRT_TCH_Lm:
+		switch (cm->chan_rate) {
+		case RSL_CMOD_SP_GSM1:
 			feature	= BTS_FEAT_SPEECH_H_V1;
 			break;
-		case GSM48_CMODE_SPEECH_AMR:
+		case RSL_CMOD_SP_GSM3:
 			feature = BTS_FEAT_SPEECH_H_AMR;
 			break;
 		default:
@@ -898,8 +863,9 @@ int bts_supports_cm(struct gsm_bts *bts, enum gsm_phys_chan_config pchan,
 		break;
 
 	default:
-		LOGP(DRSL, LOGL_ERROR, "BTS %u: unhandled pchan %s when checking mode %s\n",
-		     bts->nr, gsm_pchan_name(pchan), gsm48_chan_mode_name(cm));
+		LOGP(DRSL, LOGL_ERROR,
+		     "Unhandled RSL channel type=0x%02x/rate=0x%02x\n",
+		     cm->chan_rt, cm->chan_rate);
 		return 0;
 	}
 
@@ -930,3 +896,58 @@ struct gsm_lchan *gsm_bts_get_cbch(struct gsm_bts *bts)
 
 	return lchan;
 }
+
+/* BCCH carrier power reduction (see 3GPP TS 45.008, section 7.1) */
+int bts_set_c0_pwr_red(struct gsm_bts *bts, const uint8_t red)
+{
+	struct gsm_bts_trx *c0 = bts->c0;
+	unsigned int tn;
+
+	if (!osmo_bts_has_feature(bts->features, BTS_FEAT_BCCH_POWER_RED)) {
+		LOGPTRX(c0, DRSL, LOGL_ERROR, "BCCH carrier power reduction "
+			"is not supported by this BTS model\n");
+		return -ENOTSUP;
+	}
+
+	if (red > 6 || red % 2 != 0) {
+		LOGPTRX(c0, DRSL, LOGL_ERROR, "BCCH carrier power reduction "
+			"value (%u dB) is incorrect or out of range\n", red);
+		return -EINVAL;
+	}
+
+	LOGPTRX(c0, DRSL, LOGL_NOTICE, "BCCH carrier power reduction: "
+		"%u dB (%s)\n", red, red ? "enabled" : "disabled");
+
+	/* Timeslot 0 is always transmitting BCCH/CCCH */
+	c0->ts[0].c0_power_red_db = 0;
+
+	for (tn = 1; tn < ARRAY_SIZE(c0->ts); tn++) {
+		struct gsm_bts_trx_ts *ts = &c0->ts[tn];
+		struct gsm_bts_trx_ts *prev = ts - 1;
+
+		switch (ts_pchan(ts)) {
+		/* Not allowed on CCCH/BCCH */
+		case GSM_PCHAN_CCCH:
+			/* Preceeding timeslot shall not exceed 2 dB */
+			if (prev->c0_power_red_db > 0)
+				prev->c0_power_red_db = 2;
+			/* fall-through */
+		/* Not recommended on SDCCH/8 */
+		case GSM_PCHAN_SDCCH8_SACCH8C:
+		case GSM_PCHAN_SDCCH8_SACCH8C_CBCH:
+			ts->c0_power_red_db = 0;
+			break;
+		default:
+			ts->c0_power_red_db = red;
+			break;
+		}
+	}
+
+	/* Timeslot 7 is always preceding BCCH/CCCH */
+	if (c0->ts[7].c0_power_red_db > 0)
+		c0->ts[7].c0_power_red_db = 2;
+
+	bts->c0_power_red_db = red;
+
+	return 0;
+}
