@@ -33,7 +33,6 @@
 #include <osmocom/gsm/l1sap.h>
 #include <osmocom/gsm/gsm_utils.h>
 #include <osmocom/gsm/rsl.h>
-#include <osmocom/gsm/protocol/gsm_44_004.h>
 #include <osmocom/core/gsmtap.h>
 #include <osmocom/core/gsmtap_util.h>
 #include <osmocom/core/utils.h>
@@ -52,7 +51,6 @@
 #include <osmo-bts/abis.h>
 #include <osmo-bts/bts_model.h>
 #include <osmo-bts/handover.h>
-#include <osmo-bts/power_control.h>
 #include <osmo-bts/msg_utils.h>
 #include <osmo-bts/pcuif_proto.h>
 #include <osmo-bts/cbch.h>
@@ -95,18 +93,25 @@ unsigned int l1sap_fn2ccch_block(uint32_t fn)
 struct gsm_lchan *get_lchan_by_chan_nr(struct gsm_bts_trx *trx,
 				       unsigned int chan_nr)
 {
+	struct gsm_bts_trx_ts *ts;
 	unsigned int tn, ss;
 
 	tn = L1SAP_CHAN2TS(chan_nr);
-	OSMO_ASSERT(tn < ARRAY_SIZE(trx->ts));
+	ts = &trx->ts[tn];
+
+	if (L1SAP_IS_CHAN_VAMOS(chan_nr)) {
+		if (ts->vamos.peer == NULL)
+			return NULL;
+		ts = ts->vamos.peer;
+	}
 
 	if (L1SAP_IS_CHAN_CBCH(chan_nr))
 		ss = 2; /* CBCH is always on sub-slot 2 */
 	else
 		ss = l1sap_chan2ss(chan_nr);
-	OSMO_ASSERT(ss < ARRAY_SIZE(trx->ts[tn].lchan));
+	OSMO_ASSERT(ss < ARRAY_SIZE(ts->lchan));
 
-	return &trx->ts[tn].lchan[ss];
+	return &ts->lchan[ss];
 }
 
 static struct gsm_lchan *
@@ -132,7 +137,8 @@ static uint32_t fn_ms_adj(uint32_t fn, const struct gsm_lchan *lchan)
 		/* 12/13 frames usable for audio in TCH,
 		   160 samples per RTP packet,
 		   1 RTP packet per 4 frames */
-		samples_passed = (fn - lchan->tch.last_fn) * 12 * 160 / (13 * 4);
+		const uint32_t num_fn = GSM_TDMA_FN_SUB(fn, lchan->tch.last_fn);
+		samples_passed = num_fn * 12 * 160 / (13 * 4);
 		/* round number of samples to the nearest multiple of
 		   GSM_RTP_DURATION */
 		r = samples_passed + GSM_RTP_DURATION / 2;
@@ -466,6 +472,9 @@ static bool is_fill_frame(uint8_t chan_type, const uint8_t *data, unsigned int l
 
 	switch (chan_type) {
 	case GSMTAP_CHANNEL_AGCH:
+	case GSMTAP_CHANNEL_SDCCH:
+	case GSMTAP_CHANNEL_TCH_F:
+	case GSMTAP_CHANNEL_TCH_H:
 		if (!memcmp(data, fill_frame, GSM_MACBLOCK_LEN))
 			return true;
 		break;
@@ -473,6 +482,7 @@ static bool is_fill_frame(uint8_t chan_type, const uint8_t *data, unsigned int l
 		if (!memcmp(data, paging_fill, GSM_MACBLOCK_LEN))
 			return true;
 		break;
+	/* FIXME: implement the same for GSMTAP_CHANNEL_PDTCH from/to PCU */
 	/* don't use 'default' case here as the above only conditionally return true */
 	}
 	return false;
@@ -580,22 +590,70 @@ static unsigned int calc_exprd_rach_frames(struct gsm_bts *bts, uint32_t fn)
 	return rach_frames_expired;
 }
 
+static void l1sap_interf_meas_calc_avg(struct gsm_bts_trx *trx)
+{
+	unsigned int tn, ln;
+
+	for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+		struct gsm_bts_trx_ts *ts = &trx->ts[tn];
+
+		if (ts->mo.nm_state.operational != NM_OPSTATE_ENABLED)
+			continue;
+		if (ts->mo.nm_state.availability != NM_AVSTATE_OK)
+			continue;
+
+		for (ln = 0; ln < ARRAY_SIZE(ts->lchan); ln++) {
+			struct gsm_lchan *lchan = &ts->lchan[ln];
+
+			lchan->meas.interf_meas_avg_dbm = 0;
+			lchan->meas.interf_band = 0;
+
+			/* There must be at least one sample */
+			if (lchan->meas.interf_meas_num == 0)
+				continue;
+
+			/* Average all collected samples */
+			gsm_lchan_interf_meas_calc_avg(lchan);
+		}
+	}
+}
+
+static void l1sap_interf_meas_report(struct gsm_bts *bts)
+{
+	const uint32_t period = bts->interference.intave * 104;
+	struct gsm_bts_trx *trx;
+
+	if (bts->interference.intave == 0)
+		return;
+	if (bts->gsm_time.fn % period != 0)
+		return;
+
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		/* Calculate the average of all received samples */
+		l1sap_interf_meas_calc_avg(trx);
+		/* Report to the BSC over the A-bis/RSL */
+		rsl_tx_rf_res(trx);
+		/* Report to the PCU over the PCUIF */
+		pcu_tx_interf_ind(trx, bts->gsm_time.fn);
+	}
+}
+
 /* time information received from bts model */
 static int l1sap_info_time_ind(struct gsm_bts *bts,
 			       struct osmo_phsap_prim *l1sap,
 			       struct info_time_ind_param *info_time_ind)
 {
-	int frames_expired;
-	int i;
+	unsigned int frames_expired;
+	unsigned int i;
 
 	DEBUGPFN(DL1P, info_time_ind->fn, "Rx MPH_INFO time ind\n");
 
 	/* Calculate and check frame difference */
-	frames_expired = info_time_ind->fn - bts->gsm_time.fn;
+	frames_expired = GSM_TDMA_FN_SUB(info_time_ind->fn, bts->gsm_time.fn);
 	if (frames_expired > 1) {
 		if (bts->gsm_time.fn)
 			LOGPFN(DL1P, LOGL_ERROR, info_time_ind->fn,
-			     "Invalid condition detected: Frame difference is %"PRIu32"-%"PRIu32"=%d > 1!\n",
+			     "Invalid condition detected: Frame difference is %"PRIu32"-%"PRIu32"=%u > 1!\n",
 			     info_time_ind->fn, bts->gsm_time.fn, frames_expired);
 	}
 
@@ -612,6 +670,10 @@ static int l1sap_info_time_ind(struct gsm_bts *bts,
 		bts->load.rach.total += calc_exprd_rach_frames(bts, fn);
 	}
 
+	/* Report interference levels to the BSC */
+	if (bts_internal_flag_get(bts, BTS_INTERNAL_FLAG_INTERF_MEAS))
+		l1sap_interf_meas_report(bts);
+
 	return 0;
 }
 
@@ -636,89 +698,70 @@ static inline void set_ms_to_data(struct gsm_lchan *lchan, int16_t data, bool se
 }
 
 /* measurement information received from bts model */
-static void process_l1sap_meas_data(struct gsm_bts_trx *trx,
-				    struct osmo_phsap_prim *l1sap,
+static void process_l1sap_meas_data(struct gsm_lchan *lchan,
+				    const struct osmo_phsap_prim *l1sap,
 				    enum osmo_ph_prim ind_type)
 {
 	struct bts_ul_meas ulm;
-	struct gsm_lchan *lchan;
-	struct info_meas_ind_param *info_meas_ind;
-	struct ph_data_param *ph_data_ind;
-	struct ph_tch_param *ph_tch_ind;
-	uint8_t chan_nr;
+	const struct info_meas_ind_param *info_meas_ind;
+	const struct ph_data_param *ph_data_ind;
+	const struct ph_tch_param *ph_tch_ind;
 	uint32_t fn;
-	uint8_t inv_rssi;
-	uint8_t is_sub;
-	int16_t ta_offs_256bits;
-	uint16_t ber10k;
 	const char *ind_name;
 
 	switch (ind_type) {
 	case PRIM_MPH_INFO:
 		/* (legacy way, see also OS#2977) */
 	        info_meas_ind = &l1sap->u.info.u.meas_ind;
-		chan_nr = info_meas_ind->chan_nr;
 		fn = info_meas_ind->fn;
-		inv_rssi = info_meas_ind->inv_rssi;
-		is_sub = info_meas_ind->is_sub;
-		ta_offs_256bits = info_meas_ind->ta_offs_256bits;
-		ber10k = info_meas_ind->ber10k;
 		ind_name = "MPH INFO";
+		ulm = (struct bts_ul_meas) {
+			.ta_offs_256bits = info_meas_ind->ta_offs_256bits,
+			.inv_rssi = info_meas_ind->inv_rssi,
+			.ber10k = info_meas_ind->ber10k,
+			.c_i = info_meas_ind->c_i_cb,
+			.is_sub = info_meas_ind->is_sub,
+		};
 		break;
 	case PRIM_TCH:
 		ph_tch_ind = &l1sap->u.tch;
 		if (ph_tch_ind->rssi == 0)
 			return;
-		chan_nr = ph_tch_ind->chan_nr;
 		fn = ph_tch_ind->fn;
-		inv_rssi = abs(ph_tch_ind->rssi);
-		is_sub = ph_tch_ind->is_sub;
-		ta_offs_256bits = ph_tch_ind->ta_offs_256bits;
-		ber10k = ph_tch_ind->ber10k;
 		ind_name = "TCH";
+		ulm = (struct bts_ul_meas) {
+			.ta_offs_256bits = ph_tch_ind->ta_offs_256bits,
+			.inv_rssi = abs(ph_tch_ind->rssi),
+			.ber10k = ph_tch_ind->ber10k,
+			.c_i = ph_tch_ind->lqual_cb,
+			.is_sub = ph_tch_ind->is_sub,
+		};
 		break;
 	case PRIM_PH_DATA:
 		ph_data_ind = &l1sap->u.data;
 		if (ph_data_ind->rssi == 0)
 			return;
-		chan_nr = ph_data_ind->chan_nr;
 		fn = ph_data_ind->fn;
-		inv_rssi = abs(ph_data_ind->rssi);
-		is_sub = ph_data_ind->is_sub;
-		ta_offs_256bits = ph_data_ind->ta_offs_256bits;
-		ber10k = ph_data_ind->ber10k;
 		ind_name = "DATA";
+		ulm = (struct bts_ul_meas) {
+			.ta_offs_256bits = ph_data_ind->ta_offs_256bits,
+			.inv_rssi = abs(ph_data_ind->rssi),
+			.ber10k = ph_data_ind->ber10k,
+			.c_i = ph_data_ind->lqual_cb,
+			.is_sub = ph_data_ind->is_sub,
+		};
 		break;
 	default:
 		OSMO_ASSERT(false);
 	}
 
-	lchan = get_active_lchan_by_chan_nr(trx, chan_nr);
-	if (!lchan) {
-		LOGPFN(DL1P, LOGL_ERROR, fn,
-		       "No lchan for %s MEAS IND (chan_nr=%s)\n",
-		       ind_name, rsl_chan_nr_str(chan_nr));
-		return;
-	}
-
 	DEBUGPFN(DL1P, fn,
-		 "%s %s meas ind, ta_offs_256bits=%d, ber10k=%d, inv_rssi=%u\n",
-		 gsm_lchan_name(lchan), ind_name, ta_offs_256bits, ber10k,
-		 inv_rssi);
-
-	/* in the GPRS case we are not interested in measurement
-	 * processing.  The PCU will take care of it */
-	if (lchan->type == GSM_LCHAN_PDTCH)
-		return;
-
-	memset(&ulm, 0, sizeof(ulm));
-	ulm.ta_offs_256bits = ta_offs_256bits;
-	ulm.ber10k = ber10k;
-	ulm.inv_rssi = inv_rssi;
-	ulm.is_sub = is_sub;
+		 "%s %s meas ind, ta_offs_256bits=%d, ber10k=%d, inv_rssi=%u, C/I=%d cB\n",
+		 gsm_lchan_name(lchan), ind_name, ulm.ta_offs_256bits,
+		 ulm.ber10k, ulm.inv_rssi, ulm.c_i);
 
 	/* we assume that symbol period is 1 bit: */
-	set_ms_to_data(lchan, ta_offs_256bits / 256, true);
+	set_ms_to_data(lchan, ulm.ta_offs_256bits / 256, true);
 
 	lchan_meas_process_measurement(lchan, &ulm, fn);
 
@@ -729,6 +772,8 @@ static void process_l1sap_meas_data(struct gsm_bts_trx *trx,
 static int l1sap_mph_info_ind(struct gsm_bts_trx *trx,
 	 struct osmo_phsap_prim *l1sap, struct mph_info_param *info)
 {
+	const struct info_meas_ind_param *meas_ind;
+	struct gsm_lchan *lchan;
 	int rc = 0;
 
 	switch (info->type) {
@@ -748,7 +793,17 @@ static int l1sap_mph_info_ind(struct gsm_bts_trx *trx,
 		if (bts_internal_flag_get(trx->bts, BTS_INTERNAL_FLAG_MEAS_PAYLOAD_COMB))
 			OSMO_ASSERT(false);
 
-		process_l1sap_meas_data(trx, l1sap, PRIM_MPH_INFO);
+		meas_ind = &l1sap->u.info.u.meas_ind;
+
+		lchan = get_active_lchan_by_chan_nr(trx, meas_ind->chan_nr);
+		if (!lchan) {
+			LOGPFN(DL1P, LOGL_ERROR, meas_ind->fn,
+			       "No lchan for chan_nr=%s\n",
+			       rsl_chan_nr_str(meas_ind->chan_nr));
+			return 0;
+		}
+
+		process_l1sap_meas_data(lchan, l1sap, PRIM_MPH_INFO);
 		break;
 	default:
 		LOGP(DL1P, LOGL_NOTICE, "unknown MPH_INFO ind type %d\n",
@@ -765,6 +820,12 @@ static int l1sap_info_act_cnf(struct gsm_bts_trx *trx,
 	struct info_act_cnf_param *info_act_cnf)
 {
 	struct gsm_lchan *lchan = get_lchan_by_chan_nr(trx, info_act_cnf->chan_nr);
+	if (lchan == NULL) {
+		LOGPTRX(trx, DL1C, LOGL_ERROR, "get_lchan_by_chan_nr(chan_nr=%s) "
+			"yields NULL for PRIM_INFO_ACTIVATE.conf\n",
+			rsl_chan_nr_str(info_act_cnf->chan_nr));
+		return -ENODEV;
+	}
 
 	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "activate confirm chan_nr=%s trx=%d\n",
 		  rsl_chan_nr_str(info_act_cnf->chan_nr), trx->nr);
@@ -788,6 +849,12 @@ static int l1sap_info_rel_cnf(struct gsm_bts_trx *trx,
 	struct info_act_cnf_param *info_act_cnf)
 {
 	struct gsm_lchan *lchan = get_lchan_by_chan_nr(trx, info_act_cnf->chan_nr);
+	if (lchan == NULL) {
+		LOGPTRX(trx, DL1C, LOGL_ERROR, "get_lchan_by_chan_nr(chan_nr=%s) "
+			"yields NULL for PRIM_INFO_ACTIVATE.conf\n",
+			rsl_chan_nr_str(info_act_cnf->chan_nr));
+		return -ENODEV;
+	}
 
 	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "deactivate confirm chan_nr=%s trx=%d\n",
 		  rsl_chan_nr_str(info_act_cnf->chan_nr), trx->nr);
@@ -912,14 +979,14 @@ static inline struct msgb *lapdm_phsap_dequeue_msg_facch(struct gsm_lchan *lchan
 
 	/* Note: The repeated version of the FACCH block must be scheduled 8 or 9 bursts after the original
 	 * transmission. see 3GPP TS 44.006, section 10.2 for a more detailed explaination. */
-	if (lchan->tch.rep_facch[0].msg && GSM_TDMA_FN_SUB(fn, lchan->tch.rep_facch[0].fn) >= 8) {
+	if (lchan->rep_acch.dl_facch[0].msg && GSM_TDMA_FN_SUB(fn, lchan->rep_acch.dl_facch[0].fn) >= 8) {
 		/* Re-use stored FACCH message buffer from SLOT #0 for repetition. */
-		msg = lchan->tch.rep_facch[0].msg;
-		lchan->tch.rep_facch[0].msg = NULL;
-	} else if (lchan->tch.rep_facch[1].msg && GSM_TDMA_FN_SUB(fn, lchan->tch.rep_facch[1].fn) >= 8) {
+		msg = lchan->rep_acch.dl_facch[0].msg;
+		lchan->rep_acch.dl_facch[0].msg = NULL;
+	} else if (lchan->rep_acch.dl_facch[1].msg && GSM_TDMA_FN_SUB(fn, lchan->rep_acch.dl_facch[1].fn) >= 8) {
 		/* Re-use stored FACCH message buffer from SLOT #1 for repetition. */
-		msg = lchan->tch.rep_facch[1].msg;
-		lchan->tch.rep_facch[1].msg = NULL;
+		msg = lchan->rep_acch.dl_facch[1].msg;
+		lchan->rep_acch.dl_facch[1].msg = NULL;
 	} else {
 		/* Fetch new FACCH from queue ... */
 		if (lapdm_phsap_dequeue_prim(le, &pp) < 0)
@@ -931,16 +998,16 @@ static inline struct msgb *lapdm_phsap_dequeue_msg_facch(struct gsm_lchan *lchan
 		 * If the MS explicitly indicated that repeated ACCH is
 		 * supported, than all FACCH frames may be repeated
 		 * see also: 3GPP TS 44.006, section 10.3). */
-		if (!(lchan->repeated_acch_capability.dl_facch_all || msg->data[0] & 0x02))
+		if (!(lchan->rep_acch_cap.dl_facch_all || msg->data[0] & 0x02))
 			return msg;
 
 		/* ... and store the message buffer for repetition. */
-		if (lchan->tch.rep_facch[0].msg == NULL) {
-			lchan->tch.rep_facch[0].msg = msgb_copy(msg, "rep_facch_0");
-			lchan->tch.rep_facch[0].fn = fn;
-		} else if (lchan->tch.rep_facch[1].msg == NULL) {
-			lchan->tch.rep_facch[1].msg = msgb_copy(msg, "rep_facch_1");
-			lchan->tch.rep_facch[1].fn = fn;
+		if (lchan->rep_acch.dl_facch[0].msg == NULL) {
+			lchan->rep_acch.dl_facch[0].msg = msgb_copy(msg, "rep_facch_0");
+			lchan->rep_acch.dl_facch[0].fn = fn;
+		} else if (lchan->rep_acch.dl_facch[1].msg == NULL) {
+			lchan->rep_acch.dl_facch[1].msg = msgb_copy(msg, "rep_facch_1");
+			lchan->rep_acch.dl_facch[1].fn = fn;
 		} else {
 			/* By definition 3GPP TS 05.02 does not allow more than two (for TCH/H only one) FACCH blocks
 			 * to be transmitted simultaniously. */
@@ -951,72 +1018,6 @@ static inline struct msgb *lapdm_phsap_dequeue_msg_facch(struct gsm_lchan *lchan
 	return msg;
 }
 
-/* Decide if repeated FACCH should be applied or not. If RXQUAL level, that the
- * MS reports is high enough, FACCH repetition is not needed. */
-void repeated_dl_facch_active_decision(struct gsm_lchan *lchan, const uint8_t *l3,
-				       size_t l3_len)
-{
-	const struct gsm48_meas_res *meas_res;
-	uint8_t upper;
-	uint8_t lower;
-	uint8_t rxqual;
-
-	/* This is an optimization so that we exit as quickly as possible if
-	 * there are no FACCH repetition capabilities present. However If the
-	 * repeated FACCH capabilities vanish for whatever reason, we must be
-	 * sure that FACCH repetition is disabled. */
-	if (!lchan->repeated_acch_capability.dl_facch_cmd
-	    && !lchan->repeated_acch_capability.dl_facch_all) {
-		lchan->repeated_dl_facch_active = false;
-		return;
-	}
-
-	/* Threshold disabled (always on) */
-	if (lchan->repeated_acch_capability.rxqual == 0) {
-		lchan->repeated_dl_facch_active = true;
-		return;
-	}
-
-	/* When the MS sets the SRR bit in the UL-SACCH L1 header
-	 * (repeated SACCH requested) then it makes sense to enable
-	 * FACCH repetition too. */
-	if (lchan->meas.l1_info.srr_sro) {
-		lchan->repeated_dl_facch_active = true;
-		return;
-	}
-
-	/* Parse MS measurement results */
-	if (l3_len <= sizeof(struct gsm48_meas_res *) + 2)
-		return;
-	if (l3[0] != GSM48_PDISC_RR)
-		return;
-	if (l3[1] != GSM48_MT_RR_MEAS_REP)
-		return;
-	l3 += 2;
-	meas_res = (struct gsm48_meas_res *)l3;
-
-	/* If the RXQUAL level at the MS drops under a certain threshold
-	 * we enable FACCH repetition. */
-	upper = lchan->repeated_acch_capability.rxqual;
-	if (upper > 2)
-		lower = lchan->repeated_acch_capability.rxqual - 2;
-	else
-		lower = 0;
-
-	/* When downlink DTX is applied, use RXQUAL-SUB, otherwise use
-	 * RXQUAL-FULL. */
-	if (meas_res->dtx_used)
-		rxqual = meas_res->rxqual_sub;
-	else
-		rxqual = meas_res->rxqual_full;
-
-	if (rxqual >= upper)
-		lchan->repeated_dl_facch_active = true;
-	else if (rxqual <= lower)
-		lchan->repeated_dl_facch_active = false;
-
-}
-
 /* Special dequeueing function with SACCH repetition (3GPP TS 44.006, section 11) */
 static inline struct msgb *lapdm_phsap_dequeue_msg_sacch(struct gsm_lchan *lchan, struct lapdm_entity *le)
 {
@@ -1028,15 +1029,15 @@ static inline struct msgb *lapdm_phsap_dequeue_msg_sacch(struct gsm_lchan *lchan
 	 * possible candidates in order to have one ready in case the MS enables
 	 * SACCH repetition. */
 
-	if (lchan->rep_sacch) {
+	if (lchan->rep_acch.dl_sacch_msg) {
 		if (lchan->meas.l1_info.srr_sro == 0) {
 			/* Toss previous repetition candidate */
-			msgb_free(lchan->rep_sacch);
-			lchan->rep_sacch = NULL;
+			msgb_free(lchan->rep_acch.dl_sacch_msg);
+			lchan->rep_acch.dl_sacch_msg = NULL;
 		} else {
 			/* Use previous repetition candidate */
-			msg = lchan->rep_sacch;
-			lchan->rep_sacch = NULL;
+			msg = lchan->rep_acch.dl_sacch_msg;
+			lchan->rep_acch.dl_sacch_msg = NULL;
 			return msg;
 		}
 	}
@@ -1050,7 +1051,7 @@ static inline struct msgb *lapdm_phsap_dequeue_msg_sacch(struct gsm_lchan *lchan
 	/* Only LAPDm frames for SAPI 0 may become a repetition
 	 * candidate. */
 	if (sapi == 0)
-		lchan->rep_sacch = msgb_copy(msg, "rep_sacch");
+		lchan->rep_acch.dl_sacch_msg = msgb_copy(msg, "rep_sacch");
 
 	return msg;
 }
@@ -1130,23 +1131,41 @@ static int l1sap_ph_rts_ind(struct gsm_bts_trx *trx,
 			       rsl_chan_nr_str(chan_nr));
 			return 0;
 		}
+		if (lchan->pending_rel_ind_msg) {
+			LOGPGT(DRSL, LOGL_INFO, &g_time,
+			       "%s Forward RLL RELease INDication to the BSC\n",
+			       gsm_lchan_name(lchan));
+			abis_bts_rsl_sendmsg(lchan->pending_rel_ind_msg);
+			lchan->pending_rel_ind_msg = NULL;
+		}
 		if (L1SAP_IS_LINK_SACCH(link_id)) {
 			p = msgb_put(msg, GSM_MACBLOCK_LEN);
 			/* L1-header, if not set/modified by layer 1 */
 			p[0] = lchan->ms_power_ctrl.current;
-			if (lchan->repeated_ul_sacch_active)
+			if (lchan->rep_acch.ul_sacch_active)
 				p[0] |= 0x40; /* See also: 3GPP TS 44.004, section 7.1 */
-			p[1] = lchan->rqd_ta;
+			p[1] = lchan->ta_ctrl.current;
 			le = &lchan->lapdm_ch.lapdm_acch;
-			if (lchan->repeated_acch_capability.dl_sacch)
+			if (lchan->rep_acch_cap.dl_sacch) {
+				/* Check if MS requests SACCH repetition and update state accordingly */
+				if (lchan->meas.l1_info.srr_sro) {
+					if (lchan->rep_acch.dl_sacch_active == false)
+						LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "DL-SACCH repetition: inactive => active\n");
+					lchan->rep_acch.dl_sacch_active = true;
+				} else {
+					if (lchan->rep_acch.dl_sacch_active == true)
+						LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "DL-SACCH repetition: active => inactive\n");
+					lchan->rep_acch.dl_sacch_active = false;
+				}
 				pp_msg = lapdm_phsap_dequeue_msg_sacch(lchan, le);
-			else
+			} else {
 				pp_msg = lapdm_phsap_dequeue_msg(le);
+			}
 		} else {
 			if (lchan->ts->trx->bts->dtxd)
 				dtxd_facch = true;
 			le = &lchan->lapdm_ch.lapdm_dcch;
-			if (lchan->repeated_dl_facch_active && lchan->rsl_cmode != RSL_CMOD_SPD_SIGN)
+			if (lchan->rep_acch.dl_facch_active && lchan->rsl_cmode != RSL_CMOD_SPD_SIGN)
 				pp_msg = lapdm_phsap_dequeue_msg_facch(lchan, le, fn);
 			else
 				pp_msg = lapdm_phsap_dequeue_msg(le);
@@ -1254,7 +1273,6 @@ static int l1sap_tch_rts_ind(struct gsm_bts_trx *trx,
 	struct gsm_lchan *lchan;
 	uint8_t chan_nr, marker = 0;
 	uint32_t fn;
-	int rc;
 
 	chan_nr = rts_ind->chan_nr;
 	fn = rts_ind->fn;
@@ -1298,16 +1316,6 @@ static int l1sap_tch_rts_ind(struct gsm_bts_trx *trx,
 		resp_l1sap = msgb_l1sap_prim(resp_msg);
 	}
 
-	/* check for pending REL_IND */
-	if (lchan->pending_rel_ind_msg) {
-		LOGPGT(DRSL, LOGL_INFO, &g_time, "%s Forward REL_IND to L3\n", gsm_lchan_name(lchan));
-		/* Forward it to L3 */
-		rc = abis_bts_rsl_sendmsg(lchan->pending_rel_ind_msg);
-		lchan->pending_rel_ind_msg = NULL;
-		if (rc < 0)
-			return rc;
-	}
-
 	memset(resp_l1sap, 0, sizeof(*resp_l1sap));
 	osmo_prim_init(&resp_l1sap->oph, SAP_GSM_PH, PRIM_TCH, PRIM_OP_REQUEST,
 		resp_msg);
@@ -1407,20 +1415,21 @@ static void repeated_ul_sacch_active_decision(struct gsm_lchan *lchan,
 {
 	uint16_t upper = 0;
 	uint16_t lower = 0;
+	bool prev_repeated_ul_sacch_active = lchan->rep_acch.ul_sacch_active;
 
 	/* This is an optimization so that we exit as quickly as possible if
 	 * there are no uplink SACCH repetition capabilities present.
 	 * However If the repeated UL-SACCH capabilities vanish for whatever
 	 * reason, we must be sure that UL-SACCH repetition is disabled. */
-	if (!lchan->repeated_acch_capability.ul_sacch) {
-		lchan->repeated_ul_sacch_active = false;
-		return;
+	if (!lchan->rep_acch_cap.ul_sacch) {
+		lchan->rep_acch.ul_sacch_active = false;
+		goto out;
 	}
 
 	/* Threshold disabled (repetition is always on) */
-	if (lchan->repeated_acch_capability.rxqual == 0) {
-		lchan->repeated_ul_sacch_active = true;
-		return;
+	if (lchan->rep_acch_cap.rxqual == 0) {
+		lchan->rep_acch.ul_sacch_active = true;
+		goto out;
 	}
 
 	/* convert from RXQUAL value to ber10k value.
@@ -1433,14 +1442,22 @@ static void repeated_ul_sacch_active_decision(struct gsm_lchan *lchan,
 	 * of the table in GSM 05.08, section 8.2.4. The lower vector is just
 	 * the upper vector shifted by 2. */
 
-	upper = ber10k_by_rxqual_upper[lchan->repeated_acch_capability.rxqual];
-	lower = ber10k_by_rxqual_lower[lchan->repeated_acch_capability.rxqual];
+	upper = ber10k_by_rxqual_upper[lchan->rep_acch_cap.rxqual];
+	lower = ber10k_by_rxqual_lower[lchan->rep_acch_cap.rxqual];
 
 	/* If upper/rxqual == 0, then repeated UL-SACCH is always on */
 	if (ber10k >= upper)
-		lchan->repeated_ul_sacch_active = true;
+		lchan->rep_acch.ul_sacch_active = true;
 	else if (ber10k <= lower)
-		lchan->repeated_ul_sacch_active = false;
+		lchan->rep_acch.ul_sacch_active = false;
+
+out:
+	if (lchan->rep_acch.ul_sacch_active == prev_repeated_ul_sacch_active)
+		return;
+	if (lchan->rep_acch.ul_sacch_active)
+		LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "UL-SACCH repetition: inactive => active\n");
+	else
+		LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "UL-SACCH repetition: active => inactive\n");
 }
 
 /* DATA received from bts model */
@@ -1456,11 +1473,8 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 	uint8_t chan_nr, link_id;
 	uint8_t tn;
 	uint32_t fn;
-	int8_t rssi;
 	enum osmo_ph_pres_info_type pr_info = data_ind->pdch_presence_info;
-	struct gsm_sacch_l1_hdr *l1_hdr;
 
-	rssi = data_ind->rssi;
 	chan_nr = data_ind->chan_nr;
 	link_id = data_ind->link_id;
 	fn = data_ind->fn;
@@ -1471,21 +1485,6 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 	DEBUGPGT(DL1P, &g_time, "Rx PH-DATA.ind chan_nr=%s link_id=0x%02x len=%d\n",
 		 rsl_chan_nr_str(chan_nr), link_id, len);
 
-	/* Actually, there can be no DATA.ind on PTCCH/U (rather RACH.ind instead),
-	 * but some BTS models with buggy implementation may still be sending them
-	 * to us. Let's keep this for backwards compatibility. */
-	if (L1SAP_IS_CHAN_PDCH(chan_nr) && L1SAP_IS_PTCCH(fn)) {
-		LOGPGT(DL1P, LOGL_NOTICE, &g_time, "There can be no DATA.ind on PTCCH/U. "
-		       "This is probably a bug of the BTS model you're using, please fix!\n");
-		return -EINVAL;
-	}
-
-	/* The ph_data_param contained in the l1sap primitive may contain
-	 * measurement data. If this data is present, forward it for
-	 * processing */
-	if (bts_internal_flag_get(trx->bts, BTS_INTERNAL_FLAG_MEAS_PAYLOAD_COMB))
-		process_l1sap_meas_data(trx, l1sap, PRIM_PH_DATA);
-
 	if (ts_is_pdch(&trx->ts[tn])) {
 		lchan = get_lchan_by_chan_nr(trx, chan_nr);
 		if (!lchan)
@@ -1502,16 +1501,21 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 			return 1;
 		}
 
-		/* don't send bad frames to PCU */
-		if (len == 0)
+		/* There can be no DATA.ind on PTCCH/U (rather RACH.ind instead), but some
+		 * BTS models with buggy implementation may still be sending them to us. */
+		if (L1SAP_IS_PTCCH(fn)) {
+			LOGPGT(DL1P, LOGL_NOTICE, &g_time, "There can be no DATA.ind on PTCCH/U. "
+			       "This is probably a bug of the BTS model you're using, please fix!\n");
 			return -EINVAL;
-		/* drop incomplete UL block */
+		}
+
+		/* Drop all data from incomplete UL block */
 		if (pr_info != PRES_INFO_BOTH)
-			return 0;
+			len = 0;
 
 		/* PDTCH / PACCH frame handling */
 		pcu_tx_data_ind(&trx->ts[tn], PCU_IF_SAPI_PDTCH, fn, trx->arfcn,
-				L1SAP_FN2MACBLOCK(fn), data, len, rssi, data_ind->ber10k,
+				L1SAP_FN2MACBLOCK(fn), data, len, data_ind->rssi, data_ind->ber10k,
 				data_ind->ta_offs_256bits/64, data_ind->lqual_cb);
 		return 0;
 	}
@@ -1522,54 +1526,39 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 		return 0;
 	}
 
-	if (L1SAP_IS_LINK_SACCH(link_id))
+	/* The ph_data_param contained in the l1sap primitive may contain
+	 * measurement data. If this data is present, forward it for
+	 * processing */
+	if (bts_internal_flag_get(trx->bts, BTS_INTERNAL_FLAG_MEAS_PAYLOAD_COMB))
+		process_l1sap_meas_data(lchan, l1sap, PRIM_PH_DATA);
+
+	if (L1SAP_IS_LINK_SACCH(link_id)) {
 		repeated_ul_sacch_active_decision(lchan, data_ind->ber10k);
 
-	/* bad frame */
-	if (len == 0) {
-		if (L1SAP_IS_LINK_SACCH(link_id)) {
-			/* In case we loose a SACCH block, we must take care
-			 * that the related measurement report is sent via RSL.
-			 * This is a fallback method. The report will also
-			 * lack the measurement report from the MS side. See
-			 * also rsl.c:lapdm_rll_tx_cb() */
-			LOGPGT(DL1P, LOGL_INFO, &g_time, "Lost SACCH block, faking meas reports and ms pwr\n");
-			le = &lchan->lapdm_ch.lapdm_acch;
-			rsl_tx_meas_res(lchan, NULL, 0, le);
-
+		/* Radio Link Timeout counter */
+		if (len == 0) {
+			LOGPGT(DL1P, LOGL_INFO, &g_time, "%s Lost SACCH block\n",
+			       gsm_lchan_name(lchan));
 			radio_link_timeout(lchan, true);
-			lchan_ms_pwr_ctrl(lchan, lchan->ms_power_ctrl.current, data_ind->rssi);
+		} else {
+			radio_link_timeout(lchan, false);
 		}
-		return -EINVAL;
+
+		/* Trigger the measurement reporting/processing logic */
+		lchan_meas_handle_sacch(lchan, msg);
 	}
 
+	/* bad frame */
+	if (len == 0)
+		return -EINVAL;
+
 	/* report first valid received frame to handover process */
 	if (lchan->ho.active == HANDOVER_WAIT_FRAME)
 		handover_frame(lchan);
 
-	if (L1SAP_IS_LINK_SACCH(link_id)) {
-		radio_link_timeout(lchan, false);
+	if (L1SAP_IS_LINK_SACCH(link_id))
 		le = &lchan->lapdm_ch.lapdm_acch;
-		/* save the SACCH L1 header in the lchan struct for RSL MEAS RES */
-		if (len != GSM_MACBLOCK_LEN) {
-			LOGPGT(DL1P, LOGL_NOTICE, &g_time, "SACCH with odd len=%u!?!\n", len);
-			return -EINVAL;
-		}
-		/* Some brilliant engineer decided that the ordering of
-		 * fields on the Um interface is different from the
-		 * order of fields in RSL. See 3GPP TS 44.004 (section 7.2)
-		 * vs. 3GPP TS 48.058 (section 9.3.10). */
-		l1_hdr = (struct gsm_sacch_l1_hdr*)data;
-		lchan->meas.l1_info.ms_pwr = l1_hdr->ms_pwr;
-		lchan->meas.l1_info.fpc_epc = l1_hdr->fpc_epc;
-		lchan->meas.l1_info.srr_sro = l1_hdr->srr_sro;
-		lchan->meas.l1_info.ta = l1_hdr->ta;
-		lchan->meas.flags |= LC_UL_M_F_L1_VALID;
-
-		lchan_ms_pwr_ctrl(lchan, data[0] & 0x1f, data_ind->rssi);
-		if (trx->bts->c0 != trx) /* BS Power Control shall not be used on C0 */
-			lchan_bs_pwr_ctrl(lchan, (const struct gsm48_hdr *) &data[5]);
-	} else
+	else
 		le = &lchan->lapdm_ch.lapdm_dcch;
 
 	if (check_for_first_ciphrd(lchan, data, len))
@@ -1611,7 +1600,7 @@ static int l1sap_tch_ind(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap,
 	 * measurement data. If this data is present, forward it for
 	 * processing */
 	if (bts_internal_flag_get(trx->bts, BTS_INTERNAL_FLAG_MEAS_PAYLOAD_COMB))
-		process_l1sap_meas_data(trx, l1sap, PRIM_TCH);
+		process_l1sap_meas_data(lchan, l1sap, PRIM_TCH);
 
 	msgb_pull_to_l2(msg);
 
@@ -1880,8 +1869,12 @@ int l1sap_pdch_req(struct gsm_bts_trx_ts *ts, int is_ptcch, uint32_t fn,
 	l1sap->u.data.chan_nr = RSL_CHAN_OSMO_PDCH | ts->nr;
 	l1sap->u.data.link_id = 0x00;
 	l1sap->u.data.fn = fn;
-	msg->l2h = msgb_put(msg, len);
-	memcpy(msg->l2h, data, len);
+	if (len) {
+		msg->l2h = msgb_put(msg, len);
+		memcpy(msg->l2h, data, len);
+	} else {
+		msg->l2h = NULL; /* Idle block */
+	}
 
 	return l1sap_down(ts->trx, l1sap);
 }
@@ -1937,29 +1930,17 @@ static int l1sap_chan_act_dact_modify(struct gsm_bts_trx *trx, uint8_t chan_nr,
 int l1sap_chan_act(struct gsm_bts_trx *trx, uint8_t chan_nr, struct tlv_parsed *tp)
 {
 	struct gsm_lchan *lchan = get_lchan_by_chan_nr(trx, chan_nr);
-	struct gsm48_chan_desc *cd;
 	int rc;
 
-	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "activating channel %s\n", rsl_chan_nr_str(chan_nr));
-
-	/* osmo-pcu calls this without a valid 'tp' parameter, so we
-	 * need to make sure ew don't crash here */
-	if (tp && TLVP_PRESENT(tp, GSM48_IE_CHANDESC_2) &&
-	    TLVP_LEN(tp, GSM48_IE_CHANDESC_2) >= sizeof(*cd)) {
-		cd = (struct gsm48_chan_desc *)
-		TLVP_VAL(tp, GSM48_IE_CHANDESC_2);
-
-		/* our L1 only supports one global TSC for all channels
-		 * one one TRX, so we need to make sure not to activate
-		 * channels with a different TSC!! */
-		if (cd->h0.tsc != (lchan->ts->trx->bts->bsic & 7)) {
-			LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "lchan TSC %u != BSIC-TSC %u\n",
-				  cd->h0.tsc, lchan->ts->trx->bts->bsic & 7);
-			return -RSL_ERR_SERV_OPT_UNIMPL;
-		}
+	if (lchan->state == LCHAN_S_ACTIVE) {
+		LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "Trying to activate already active channel %s\n",
+			  rsl_chan_nr_str(chan_nr));
+		return -1;
 	}
 
-	lchan->s = lchan->ts->trx->bts->radio_link_timeout.current;
+	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "Activating channel %s\n", rsl_chan_nr_str(chan_nr));
+
+	lchan->s = trx->bts->radio_link_timeout.current;
 
 	rc = l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_ACTIVATE, 0);
 	if (rc)
@@ -1967,18 +1948,22 @@ int l1sap_chan_act(struct gsm_bts_trx *trx, uint8_t chan_nr, struct tlv_parsed *
 
 	/* Init DTX DL FSM if necessary */
 	if (trx->bts->dtxd && lchan->type != GSM_LCHAN_SDCCH) {
-		char name[32];
-		snprintf(name, sizeof(name), "bts%u-trx%u-ts%u-ss%u", lchan->ts->trx->bts->nr,
-			 lchan->ts->trx->nr, lchan->ts->nr, lchan->nr);
 		lchan->tch.dtx.dl_amr_fsm = osmo_fsm_inst_alloc(&dtx_dl_amr_fsm,
 								tall_bts_ctx,
 								lchan,
 								LOGL_DEBUG,
-								name);
+								NULL);
 		if (!lchan->tch.dtx.dl_amr_fsm) {
 			l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_DEACTIVATE, 0);
 			return -RSL_ERR_EQUIPMENT_FAIL;
 		}
+
+		rc = osmo_fsm_inst_update_id_f(lchan->tch.dtx.dl_amr_fsm,
+					       "bts%u-trx%u-ts%u-ss%u%s",
+					       trx->bts->nr, trx->nr,
+					       lchan->ts->nr, lchan->nr,
+					       lchan->ts->vamos.is_shadow ? "-shadow" : "");
+		OSMO_ASSERT(rc == 0);
 	}
 	return 0;
 }
@@ -1986,8 +1971,15 @@ int l1sap_chan_act(struct gsm_bts_trx *trx, uint8_t chan_nr, struct tlv_parsed *
 int l1sap_chan_rel(struct gsm_bts_trx *trx, uint8_t chan_nr)
 {
 	struct gsm_lchan *lchan = get_lchan_by_chan_nr(trx, chan_nr);
-	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "deactivating channel chan_nr=%s trx=%d\n",
-		  rsl_chan_nr_str(chan_nr), trx->nr);
+
+	if (lchan->state == LCHAN_S_NONE) {
+		LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "Trying to deactivate already deactivated channel %s\n",
+			  rsl_chan_nr_str(chan_nr));
+		return -1;
+	}
+
+	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "Deactivating channel %s\n",
+		  rsl_chan_nr_str(chan_nr));
 
 	if (lchan->tch.dtx.dl_amr_fsm) {
 		osmo_fsm_inst_free(lchan->tch.dtx.dl_amr_fsm);
@@ -2002,8 +1994,8 @@ int l1sap_chan_deact_sacch(struct gsm_bts_trx *trx, uint8_t chan_nr)
 {
 	struct gsm_lchan *lchan = get_lchan_by_chan_nr(trx, chan_nr);
 
-	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "deactivating sacch chan_nr=%s trx=%d\n",
-		  rsl_chan_nr_str(chan_nr), trx->nr);
+	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "Deactivating SACCH on channel %s\n",
+		  rsl_chan_nr_str(chan_nr));
 
 	return l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_DEACTIVATE,
 		1);
@@ -2011,8 +2003,10 @@ int l1sap_chan_deact_sacch(struct gsm_bts_trx *trx, uint8_t chan_nr)
 
 int l1sap_chan_modify(struct gsm_bts_trx *trx, uint8_t chan_nr)
 {
-	LOGP(DL1C, LOGL_INFO, "modifying channel chan_nr=%s trx=%d\n",
-		rsl_chan_nr_str(chan_nr), trx->nr);
+	struct gsm_lchan *lchan = get_lchan_by_chan_nr(trx, chan_nr);
+
+	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "Modifying channel %s\n",
+		rsl_chan_nr_str(chan_nr));
 
 	return l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_MODIFY, 0);
 }
