@@ -2,14 +2,18 @@
 #include <stdint.h>
 #include <errno.h>
 
-#include <osmocom/gsm/gsm_utils.h>
 #include <osmocom/core/utils.h>
+#include <osmocom/core/endian.h>
+
+#include <osmocom/gsm/gsm_utils.h>
+#include <osmocom/gsm/protocol/gsm_44_004.h>
 
 #include <osmo-bts/gsm_data.h>
 #include <osmo-bts/logging.h>
 #include <osmo-bts/measurement.h>
 #include <osmo-bts/scheduler.h>
 #include <osmo-bts/rsl.h>
+#include <osmo-bts/power_control.h>
 #include <osmo-bts/ta_control.h>
 
 /* Tables as per TS 45.008 Section 8.3 */
@@ -250,11 +254,6 @@ int is_meas_complete(struct gsm_lchan *lchan, uint32_t fn)
 	int rc = 0;
 	enum gsm_phys_chan_config pchan = ts_pchan(lchan->ts);
 
-	if (lchan->ts->nr >= 8)
-		return -EINVAL;
-	if (pchan >= _GSM_PCHAN_MAX)
-		return -EINVAL;
-
 	switch (pchan) {
 	case GSM_PCHAN_TCH_F:
 		fn_mod = translate_tch_meas_rep_fn104(fn % 104);
@@ -347,7 +346,7 @@ int lchan_new_ul_meas(struct gsm_lchan *lchan,
 	if (!ulm->is_sub)
 		dest->is_sub = ts45008_83_is_sub(lchan, fn);
 
-	DEBUGPFN(DMEAS, fn, "%s adding measurement (ber10k=%u, ta_offs=%d, ci=%0.2f, is_sub=%u, rssi=-%u), num_ul_meas=%d, fn_mod=%u\n",
+	DEBUGPFN(DMEAS, fn, "%s adding measurement (ber10k=%u, ta_offs=%d, ci_cB=%d, is_sub=%u, rssi=-%u), num_ul_meas=%d, fn_mod=%u\n",
 		 gsm_lchan_name(lchan), ulm->ber10k, ulm->ta_offs_256bits,
 		 ulm->c_i, dest->is_sub, ulm->inv_rssi, lchan->meas.num_ul_meas,
 		 fn_mod);
@@ -561,8 +560,10 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 	struct gsm_meas_rep_unidir *mru;
 	uint32_t ber_full_sum = 0;
 	uint32_t irssi_full_sum = 0;
+	int32_t ci_full_sum = 0;
 	uint32_t ber_sub_sum = 0;
 	uint32_t irssi_sub_sum = 0;
+	int32_t ci_sub_sum = 0;
 	int32_t ta256b_sum = 0;
 	unsigned int num_meas_sub = 0;
 	unsigned int num_meas_sub_actual = 0;
@@ -594,7 +595,7 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 		/* When AMR is used, we expect at least one SUB frame, since
 		 * the SACCH will always be SUB frame. There may occur more
 		 * SUB frames but since DTX periods in AMR are dynamic, we
-		 * can not know how much exactly. */
+		 * can not know how many exactly. */
 		num_meas_sub_expect = 1;
 	}
 
@@ -630,11 +631,13 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 			m = &lchan->meas.uplink[i + num_ul_meas_excess];
 			if (m->is_sub) {
 				irssi_sub_sum += m->inv_rssi;
+				ci_sub_sum += m->c_i;
 				num_meas_sub_actual++;
 				is_sub = true;
 			}
 			irssi_full_sum += m->inv_rssi;
 			ta256b_sum += m->ta_offs_256bits;
+			ci_full_sum += m->c_i;
 
 			num_ul_meas_actual++;
 		} else {
@@ -703,27 +706,32 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 	else
 		irssi_full_sum = irssi_full_sum / num_ul_meas_actual;
 
-	if (!num_ul_meas_actual)
+	if (!num_ul_meas_actual) {
 		ta256b_sum = lchan->meas.ms_toa256;
-	else
+		ci_full_sum = lchan->meas.ul_ci_cb_full;
+	} else {
 		ta256b_sum = ta256b_sum / (signed)num_ul_meas_actual;
+		ci_full_sum = ci_full_sum / (signed)num_ul_meas_actual;
+	}
 
 	if (!num_meas_sub)
 		ber_sub_sum = MEASUREMENT_DUMMY_BER;
 	else
 		ber_sub_sum = ber_sub_sum / num_meas_sub;
 
-	if (!num_meas_sub_actual)
+	if (!num_meas_sub_actual) {
 		irssi_sub_sum = MEASUREMENT_DUMMY_IRSSI;
-	else
+		ci_sub_sum = lchan->meas.ul_ci_cb_sub;
+	} else {
 		irssi_sub_sum = irssi_sub_sum / num_meas_sub_actual;
+		ci_sub_sum = ci_sub_sum / (signed)num_meas_sub_actual;
+	}
 
 	LOGPLCHAN(lchan, DMEAS, LOGL_INFO,
-		  "Computed TA256(% 4d) BER-FULL(%2u.%02u%%), RSSI-FULL(-%3udBm), "
-		  "BER-SUB(%2u.%02u%%), RSSI-SUB(-%3udBm)\n",
-		  ta256b_sum, ber_full_sum / 100, ber_full_sum % 100,
-		  irssi_full_sum, ber_sub_sum / 100, ber_sub_sum % 100,
-		  irssi_sub_sum);
+		  "Computed TA256(% 4d), BER-FULL(%2u.%02u%%), RSSI-FULL(-%3udBm), C/I-FULL(% 4d cB), "
+		  "BER-SUB(%2u.%02u%%), RSSI-SUB(-%3udBm), C/I-SUB(% 4d cB)\n",
+		  ta256b_sum, ber_full_sum / 100, ber_full_sum % 100, irssi_full_sum, ci_full_sum,
+		  ber_sub_sum / 100, ber_sub_sum % 100, irssi_sub_sum, ci_sub_sum);
 
 	/* store results */
 	mru = &lchan->meas.ul_res;
@@ -732,6 +740,8 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 	mru->full.rx_qual = ber10k_to_rxqual(ber_full_sum);
 	mru->sub.rx_qual = ber10k_to_rxqual(ber_sub_sum);
 	lchan->meas.ms_toa256 = ta256b_sum;
+	lchan->meas.ul_ci_cb_full = ci_full_sum;
+	lchan->meas.ul_ci_cb_sub = ci_sub_sum;
 
 	LOGPLCHAN(lchan, DMEAS, LOGL_INFO,
 		  "UL MEAS RXLEV_FULL(%u), RXLEV_SUB(%u), RXQUAL_FULL(%u), RXQUAL_SUB(%u), "
@@ -744,11 +754,6 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 
 	lchan_meas_compute_extended(lchan);
 
-	/* Compute new ta_req value. This has to be done here since the value
-	 * in lchan->meas.num_ul_meas together with lchan->meas.ms_toa256
-	 * is needed for the computation. */
-	lchan_ms_ta_ctrl(lchan);
-
 	lchan->meas.num_ul_meas = 0;
 
 	/* return 1 to indicate that the computation has been done and the next
@@ -776,3 +781,226 @@ void lchan_meas_reset(struct gsm_lchan *lchan)
 	memset(&lchan->meas, 0, sizeof(lchan->meas));
 	lchan->meas.last_fn = LCHAN_FN_DUMMY;
 }
+
+static inline uint8_t ms_to2rsl(const struct gsm_lchan *lchan, uint8_t ta)
+{
+	return (lchan->ms_t_offs >= 0) ? lchan->ms_t_offs : (lchan->p_offs - ta);
+}
+
+static inline bool ms_to_valid(const struct gsm_lchan *lchan)
+{
+	return (lchan->ms_t_offs >= 0) || (lchan->p_offs >= 0);
+}
+
+/* Decide if repeated FACCH should be applied or not. If RXQUAL level, that the
+ * MS reports is high enough, FACCH repetition is not needed. */
+static void repeated_dl_facch_active_decision(struct gsm_lchan *lchan,
+					      const struct gsm48_meas_res *meas_res)
+{
+	uint8_t upper;
+	uint8_t lower;
+	uint8_t rxqual;
+	bool prev_repeated_dl_facch_active = lchan->rep_acch.dl_facch_active;
+
+	/* This is an optimization so that we exit as quickly as possible if
+	 * there are no FACCH repetition capabilities present. However If the
+	 * repeated FACCH capabilities vanish for whatever reason, we must be
+	 * sure that FACCH repetition is disabled. */
+	if (!lchan->rep_acch_cap.dl_facch_cmd
+	    && !lchan->rep_acch_cap.dl_facch_all) {
+		lchan->rep_acch.dl_facch_active = false;
+		goto out;
+	}
+
+	/* Threshold disabled (always on) */
+	if (lchan->rep_acch_cap.rxqual == 0) {
+		lchan->rep_acch.dl_facch_active = true;
+		goto out;
+	}
+
+	/* When the MS sets the SRR bit in the UL-SACCH L1 header
+	 * (repeated SACCH requested) then it makes sense to enable
+	 * FACCH repetition too. */
+	if (lchan->meas.l1_info.srr_sro) {
+		lchan->rep_acch.dl_facch_active = true;
+		goto out;
+	}
+
+	/* Parse MS measurement results */
+	if (meas_res == NULL)
+		goto out;
+	if (!gsm48_meas_res_is_valid(meas_res))
+		goto out;
+
+	/* If the RXQUAL level at the MS drops under a certain threshold
+	 * we enable FACCH repetition. */
+	upper = lchan->rep_acch_cap.rxqual;
+	if (upper > 2)
+		lower = lchan->rep_acch_cap.rxqual - 2;
+	else
+		lower = 0;
+
+	/* When downlink DTX is applied, use RXQUAL-SUB, otherwise use
+	 * RXQUAL-FULL. */
+	if (meas_res->dtx_used)
+		rxqual = meas_res->rxqual_sub;
+	else
+		rxqual = meas_res->rxqual_full;
+
+	if (rxqual >= upper)
+		lchan->rep_acch.dl_facch_active = true;
+	else if (rxqual <= lower)
+		lchan->rep_acch.dl_facch_active = false;
+
+out:
+	if (lchan->rep_acch.dl_facch_active == prev_repeated_dl_facch_active)
+		return;
+	if (lchan->rep_acch.dl_facch_active)
+		LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "DL-FACCH repetition: inactive => active\n");
+	else
+		LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "DL-FACCH repetition: active => inactive\n");
+}
+
+static void acch_overpower_active_decision(struct gsm_lchan *lchan,
+					   const struct gsm48_meas_res *meas_res)
+{
+	const bool old = lchan->top_acch_active;
+	uint8_t upper, lower, rxqual;
+
+	/* ACCH overpower is not allowed => nothing to do */
+	if (lchan->top_acch_cap.overpower_db == 0)
+		return;
+	/* RxQual threshold is disabled => overpower is always on */
+	if (lchan->top_acch_cap.rxqual == 0)
+		return;
+
+	/* If DTx is active on Downlink, use the '-SUB' */
+	if (meas_res->dtx_used)
+		rxqual = meas_res->rxqual_sub;
+	else /* ... otherwise use the '-FULL' */
+		rxqual = meas_res->rxqual_full;
+
+	upper = lchan->top_acch_cap.rxqual;
+	if (upper > 2)
+		lower = upper - 2;
+	else
+		lower = 0;
+
+	if (rxqual >= upper)
+		lchan->top_acch_active = true;
+	else if (rxqual <= lower)
+		lchan->top_acch_active = false;
+
+	if (lchan->top_acch_active != old) {
+		LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "Temporary ACCH overpower: %s\n",
+			  lchan->top_acch_active ? "inactive => active"
+						  : "active => inactive");
+	}
+}
+
+static bool data_is_rr_meas_rep(const uint8_t *data)
+{
+	const struct gsm48_hdr *gh = (void *)(data + 5);
+	const uint8_t *lapdm_hdr = (void *)(data + 2);
+
+	/* LAPDm address field: SAPI=0, C/R=0, EA=1 */
+	if (lapdm_hdr[0] != 0x01)
+		return false;
+	/* LAPDm control field: U, func=UI */
+	if (lapdm_hdr[1] != 0x03)
+		return false;
+	/* Protocol discriminator: RR */
+	if (gh->proto_discr != GSM48_PDISC_RR)
+		return false;
+
+	switch (gh->msg_type) {
+	case GSM48_MT_RR_EXT_MEAS_REP:
+	case GSM48_MT_RR_MEAS_REP:
+		return true;
+	default:
+		return false;
+	}
+}
+
+/* Called every time a SACCH block is received from lower layers */
+void lchan_meas_handle_sacch(struct gsm_lchan *lchan, struct msgb *msg)
+{
+	const struct gsm48_meas_res *mr = NULL;
+	const struct gsm48_hdr *gh = NULL;
+	int timing_offset, rc;
+	bool dtxu_used = true; /* safe default assumption */
+	uint8_t ms_pwr;
+	uint8_t ms_ta;
+	int8_t ul_rssi;
+	int16_t ul_ci_cb;
+
+	if (msgb_l2len(msg) == GSM_MACBLOCK_LEN) {
+		/* Some brilliant engineer decided that the ordering of
+		 * fields on the Um interface is different from the
+		 * order of fields in RSL. See 3GPP TS 44.004 (section 7.2)
+		 * vs. 3GPP TS 48.058 (section 9.3.10). */
+		const struct gsm_sacch_l1_hdr *l1h = msgb_l2(msg);
+		lchan->meas.l1_info.ms_pwr = l1h->ms_pwr;
+		lchan->meas.l1_info.fpc_epc = l1h->fpc_epc;
+		lchan->meas.l1_info.srr_sro = l1h->srr_sro;
+		lchan->meas.l1_info.ta = l1h->ta;
+		lchan->meas.flags |= LC_UL_M_F_L1_VALID;
+
+		/* Check if this is a Measurement Report */
+		if (data_is_rr_meas_rep(msgb_l2(msg))) {
+			/* Skip both L1 SACCH and LAPDm headers */
+			msg->l3h = (void *)(msg->l2h + 2 + 3);
+			gh = msgb_l3(msg);
+		}
+
+		ms_pwr = lchan->meas.l1_info.ms_pwr;
+		ms_ta = lchan->meas.l1_info.ta;
+	} else {
+		lchan->meas.flags &= ~LC_UL_M_F_L1_VALID;
+		ms_pwr = lchan->ms_power_ctrl.current;
+		ms_ta = lchan->ta_ctrl.current;
+	}
+
+	timing_offset = ms_to_valid(lchan) ? ms_to2rsl(lchan, ms_ta) : -1;
+	rc = rsl_tx_meas_res(lchan, msgb_l3(msg), msgb_l3len(msg), timing_offset);
+	if (rc == 0) /* Count successful transmissions */
+		lchan->meas.res_nr++;
+
+	/* Run control loops now that we have all the information: */
+	/* 3GPP TS 45.008 sec 4.2: UL L1 SACCH Header contains TA and
+	 * MS_PWR used "for the last burst of the previous SACCH
+	 * period". Since MS must use the values provided in DL SACCH
+	 * starting at next meas period, the value of the "last burst"
+	 * is actually the value used in the entire meas period. Since
+	 * it contains info about the previous meas period, we want to
+	 * feed the Control Loop with the measurements for the same
+	 * period (the previous one), which is stored in lchan->meas(.ul_res):
+	 */
+	if (gh && gh->msg_type == GSM48_MT_RR_MEAS_REP) {
+		mr = (const struct gsm48_meas_res *)gh->data;
+		if (gsm48_meas_res_is_valid(mr))
+			dtxu_used = mr->dtx_used;
+	}
+
+	if (dtxu_used) {
+		ul_rssi = rxlev2dbm(lchan->meas.ul_res.sub.rx_lev);
+		ul_ci_cb = lchan->meas.ul_ci_cb_sub;
+	} else {
+		ul_rssi = rxlev2dbm(lchan->meas.ul_res.full.rx_lev);
+		ul_ci_cb = lchan->meas.ul_ci_cb_full;
+	}
+	lchan_ms_ta_ctrl(lchan, ms_ta, lchan->meas.ms_toa256);
+	lchan_ms_pwr_ctrl(lchan, ms_pwr, ul_rssi, ul_ci_cb);
+	if (mr && gsm48_meas_res_is_valid(mr)) {
+		lchan_bs_pwr_ctrl(lchan, mr);
+		acch_overpower_active_decision(lchan, mr);
+	}
+
+	repeated_dl_facch_active_decision(lchan, mr);
+
+	/* Reset state for next iteration */
+	lchan->tch.dtx.dl_active = false;
+	lchan->meas.flags &= ~LC_UL_M_F_OSMO_EXT_VALID;
+	lchan->ms_t_offs = -1;
+	lchan->p_offs = -1;
+}
