@@ -340,12 +340,13 @@ void oml_mo_state_init(struct gsm_abis_mo *mo, int op_state, int avail_state)
 	mo->nm_state.operational = op_state;
 }
 
-int oml_mo_state_chg(struct gsm_abis_mo *mo, int op_state, int avail_state)
+int oml_mo_state_chg(struct gsm_abis_mo *mo, int op_state, int avail_state, int adm_state)
 {
 	int rc = 0;
 
 	if ((op_state != -1 && mo->nm_state.operational != op_state) ||
-	    (avail_state != -1 && mo->nm_state.availability != avail_state)) {
+	    (avail_state != -1 && mo->nm_state.availability != avail_state) ||
+	    (adm_state != -1 && mo->nm_state.administrative != adm_state)) {
 		if (avail_state != -1) {
 			LOGP(DOML, LOGL_INFO, "%s AVAIL STATE %s -> %s\n",
 				gsm_abis_mo_name(mo),
@@ -354,13 +355,25 @@ int oml_mo_state_chg(struct gsm_abis_mo *mo, int op_state, int avail_state)
 			mo->nm_state.availability = avail_state;
 		}
 		if (op_state != -1) {
+			struct nm_statechg_signal_data nsd;
 			LOGP(DOML, LOGL_INFO, "%s OPER STATE %s -> %s\n",
 				gsm_abis_mo_name(mo),
 				abis_nm_opstate_name(mo->nm_state.operational),
 				abis_nm_opstate_name(op_state));
+			nsd.mo = mo;
+			nsd.old_state = mo->nm_state.operational;
+			nsd.new_state = op_state;
 			mo->nm_state.operational = op_state;
-			osmo_signal_dispatch(SS_GLOBAL, S_NEW_OP_STATE, NULL);
+			osmo_signal_dispatch(SS_GLOBAL, S_NEW_OP_STATE, &nsd);
 		}
+		if (adm_state != -1) {
+			LOGP(DOML, LOGL_INFO, "%s ADMIN STATE %s -> %s\n",
+				gsm_abis_mo_name(mo),
+				abis_nm_admin_name(mo->nm_state.administrative),
+				abis_nm_admin_name(adm_state));
+			mo->nm_state.administrative = adm_state;
+		}
+
 
 		/* send state change report */
 		rc = oml_tx_state_changed(mo);
@@ -585,8 +598,8 @@ static int oml_rx_set_bts_attr(struct gsm_bts *bts, struct msgb *msg)
 	/* 9.4.25 Interference Level Boundaries */
 	if (TLVP_PRES_LEN(&tp, NM_ATT_INTERF_BOUND, 6)) {
 		payload = TLVP_VAL(&tp, NM_ATT_INTERF_BOUND);
-		for (i = 0; i < 6; i++) {
-			int16_t boundary = *payload;
+		for (i = 0; i < ARRAY_SIZE(bts->interference.boundary); i++) {
+			const int16_t boundary = payload[i];
 			bts->interference.boundary[i] = -1 * boundary;
 		}
 	}
@@ -661,7 +674,10 @@ static int oml_rx_set_bts_attr(struct gsm_bts *bts, struct msgb *msg)
 	/* 9.4.11 CCCH Load Indication Period */
 	if (TLVP_PRES_LEN(&tp, NM_ATT_CCCH_L_I_P, 1)) {
 		bts->load.ccch.load_ind_period = *TLVP_VAL(&tp, NM_ATT_CCCH_L_I_P);
-		load_timer_start(bts);
+		if (load_timer_is_running(bts)) {
+			load_timer_stop(bts);
+			load_timer_start(bts);
+		}
 	}
 
 	/* 9.4.44 RACH Busy Threshold */
@@ -789,9 +805,12 @@ static int oml_rx_set_radio_attr(struct gsm_bts_trx *trx, struct msgb *msg)
 	return bts_model_apply_oml(trx->bts, msg, tp_merged, NM_OC_RADIO_CARRIER, trx);
 }
 
-static int conf_lchans(struct gsm_bts_trx_ts *ts)
+static int handle_chan_comb(struct gsm_bts_trx_ts *ts, const uint8_t comb)
 {
-	enum gsm_phys_chan_config pchan = ts->pchan;
+	enum gsm_phys_chan_config pchan;
+
+	pchan = abis_nm_pchan4chcomb(comb);
+	ts->pchan = pchan;
 
 	/* RSL_MT_IPAC_PDCH_ACT style dyn PDCH */
 	if (pchan == GSM_PCHAN_TCH_F_PDCH)
@@ -799,7 +818,7 @@ static int conf_lchans(struct gsm_bts_trx_ts *ts)
 						    : GSM_PCHAN_TCH_F;
 
 	/* Osmocom RSL CHAN ACT style dyn TS */
-	if (pchan == GSM_PCHAN_TCH_F_TCH_H_PDCH) {
+	if (pchan == GSM_PCHAN_OSMO_DYN) {
 		pchan = ts->dyn.pchan_is;
 
 		/* If the dyn TS doesn't have a pchan yet, do nothing. */
@@ -810,62 +829,65 @@ static int conf_lchans(struct gsm_bts_trx_ts *ts)
 	return conf_lchans_as_pchan(ts, pchan);
 }
 
+static inline void lchans_type_set(struct gsm_bts_trx_ts *ts,
+				   enum gsm_chan_t lchan_type,
+				   unsigned int num_lchans)
+{
+	unsigned int i;
+
+	for (i = 0; i < num_lchans; i++)
+		ts->lchan[i].type = lchan_type;
+}
+
 int conf_lchans_as_pchan(struct gsm_bts_trx_ts *ts,
 			 enum gsm_phys_chan_config pchan)
 {
-	struct gsm_lchan *lchan;
-	unsigned int i;
+	/* Initialize all lchans with GSM_LCHAN_NONE first */
+	lchans_type_set(ts, GSM_LCHAN_NONE, ARRAY_SIZE(ts->lchan));
 
 	switch (pchan) {
 	case GSM_PCHAN_CCCH_SDCCH4_CBCH:
-		/* fallthrough */
 	case GSM_PCHAN_CCCH_SDCCH4:
-		for (i = 0; i < 4; i++) {
-			lchan = &ts->lchan[i];
-			if (pchan == GSM_PCHAN_CCCH_SDCCH4_CBCH
-			    && i == 2) {
-				lchan->type = GSM_LCHAN_CBCH;
-			} else {
-				lchan->type = GSM_LCHAN_SDCCH;
-			}
-		}
+		lchans_type_set(ts, GSM_LCHAN_SDCCH, 4);
+		if (pchan == GSM_PCHAN_CCCH_SDCCH4_CBCH)
+			ts->lchan[2].type = GSM_LCHAN_CBCH;
 		/* fallthrough */
 	case GSM_PCHAN_CCCH:
-		lchan = &ts->lchan[CCCH_LCHAN];
-		lchan->type = GSM_LCHAN_CCCH;
+		ts->lchan[CCCH_LCHAN].type = GSM_LCHAN_CCCH;
 		break;
 	case GSM_PCHAN_TCH_F:
-		lchan = &ts->lchan[0];
-		lchan->type = GSM_LCHAN_TCH_F;
+		if (ts->vamos.peer != NULL) { /* VAMOS: enable shadow lchans */
+			lchans_type_set(ts->vamos.peer, GSM_LCHAN_TCH_F, 1);
+			ts->vamos.peer->pchan = GSM_PCHAN_TCH_F;
+		}
+		lchans_type_set(ts, GSM_LCHAN_TCH_F, 1);
 		break;
 	case GSM_PCHAN_TCH_H:
-		for (i = 0; i < 2; i++) {
-			lchan = &ts->lchan[i];
-			lchan->type = GSM_LCHAN_TCH_H;
+		if (ts->vamos.peer != NULL) { /* VAMOS: enable shadow lchans */
+			lchans_type_set(ts->vamos.peer, GSM_LCHAN_TCH_H, 2);
+			ts->vamos.peer->pchan = GSM_PCHAN_TCH_H;
 		}
+		lchans_type_set(ts, GSM_LCHAN_TCH_H, 2);
 		break;
 	case GSM_PCHAN_SDCCH8_SACCH8C_CBCH:
-		/* fallthrough */
 	case GSM_PCHAN_SDCCH8_SACCH8C:
-		for (i = 0; i < 8; i++) {
-			lchan = &ts->lchan[i];
-			if (pchan == GSM_PCHAN_SDCCH8_SACCH8C_CBCH
-			    && i == 2) {
-				lchan->type = GSM_LCHAN_CBCH;
-			} else {
-				lchan->type = GSM_LCHAN_SDCCH;
-			}
-		}
+		lchans_type_set(ts, GSM_LCHAN_SDCCH, 8);
+		if (pchan == GSM_PCHAN_SDCCH8_SACCH8C_CBCH)
+			ts->lchan[2].type = GSM_LCHAN_CBCH;
 		break;
 	case GSM_PCHAN_PDCH:
-		lchan = &ts->lchan[0];
-		lchan->type = GSM_LCHAN_PDTCH;
+		if (ts->vamos.peer != NULL) { /* VAMOS: disable shadow lchans */
+			lchans_type_set(ts->vamos.peer, GSM_LCHAN_NONE, 1);
+			ts->vamos.peer->pchan = GSM_PCHAN_NONE;
+		}
+		lchans_type_set(ts, GSM_LCHAN_PDTCH, 1);
 		break;
 	default:
 		LOGP(DOML, LOGL_ERROR, "Unknown/unhandled PCHAN type: %u %s\n",
 		     ts->pchan, gsm_pchan_name(ts->pchan));
 		return -NM_NACK_PARAM_RANGE;
 	}
+
 	return 0;
 }
 
@@ -946,12 +968,10 @@ static int oml_rx_set_chan_attr(struct gsm_bts_trx_ts *ts, struct msgb *msg)
 
 	/* 9.4.13 Channel Combination */
 	if (TLVP_PRES_LEN(&tp, NM_ATT_CHAN_COMB, 1)) {
-		uint8_t comb = *TLVP_VAL(&tp, NM_ATT_CHAN_COMB);
-		ts->pchan = abis_nm_pchan4chcomb(comb);
-		rc = conf_lchans(ts);
-		if (rc < 0) {
+		const uint8_t comb = *TLVP_VAL(&tp, NM_ATT_CHAN_COMB);
+		if ((rc = handle_chan_comb(ts, comb)) != 0) {
 			LOGPFOH(DOML, LOGL_ERROR, foh, "SET CHAN ATTR: invalid Chan Comb 0x%x"
-				" (pchan=%s, conf_lchans()->%d)\n",
+				" (pchan=%s, handle_chan_comb() returns %d)\n",
 				comb, gsm_pchan_name(ts->pchan), rc);
 			talloc_free(tp_merged);
 			/* Send NACK */
@@ -961,10 +981,17 @@ static int oml_rx_set_chan_attr(struct gsm_bts_trx_ts *ts, struct msgb *msg)
 
 	/* 9.4.60 TSC */
 	if (TLVP_PRES_LEN(&tp, NM_ATT_TSC, 1)) {
-		ts->tsc = *TLVP_VAL(&tp, NM_ATT_TSC);
+		ts->tsc_oml = ts->tsc = *TLVP_VAL(&tp, NM_ATT_TSC);
+		if (ts->tsc != BTS_TSC(bts) &&
+		    !osmo_bts_has_feature(bts->features, BTS_FEAT_MULTI_TSC)) {
+			LOGPFOH(DOML, LOGL_ERROR, foh, "SET CHAN ATTR: this BTS model does not "
+				"support TSC %u != BSIC-BCC %u\n", ts->tsc, BTS_TSC(bts));
+			talloc_free(tp_merged);
+			return oml_fom_ack_nack(msg, NM_NACK_PARAM_RANGE);
+		}
 	} else {
 		/* If there is no TSC specified, use the BCC */
-		ts->tsc = BSIC2BCC(bts->bsic);
+		ts->tsc_oml = ts->tsc = BTS_TSC(bts);
 	}
 	LOGPFOH(DOML, LOGL_INFO, foh, "SET CHAN ATTR (TSC=%u pchan=%s",
 		ts->tsc, gsm_pchan_name(ts->pchan));
