@@ -682,7 +682,7 @@ static enum gsm_phys_chan_config pick_pchan(struct gsm_bts_trx_ts *ts)
 		if (ts->flags & TS_F_PDCH_ACTIVE)
 			return GSM_PCHAN_PDCH;
 		return GSM_PCHAN_TCH_F;
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
 		return ts->dyn.pchan_is;
 	default:
 		return ts->pchan;
@@ -696,7 +696,7 @@ static uint8_t chan_nr_by_sapi(struct gsm_bts_trx_ts *ts,
 	uint8_t cbits = 0;
 	enum gsm_phys_chan_config pchan = pick_pchan(ts);
 	OSMO_ASSERT(pchan != GSM_PCHAN_TCH_F_PDCH);
-	OSMO_ASSERT(pchan != GSM_PCHAN_TCH_F_TCH_H_PDCH);
+	OSMO_ASSERT(pchan != GSM_PCHAN_OSMO_DYN);
 
 	switch (sapi) {
 	case GsmL1_Sapi_Bcch:
