@@ -1,6 +1,7 @@
 /*
  * (C) 2013 by Andreas Eversberg <jolly@eversberg.eu>
  * (C) 2015-2017 by Harald Welte <laforge@gnumonks.org>
+ * Contributions by sysmocom - s.f.m.c. GmbH
  *
  * All Rights Reserved
  *
@@ -45,11 +46,9 @@
 #include <loops.h>
 
 /*! \brief a single TCH/F burst was received by the PHY, process it */
-int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
-	       uint8_t bid, const struct trx_ul_burst_ind *bi)
+int rx_tchf_fn(struct l1sched_ts *l1ts, const struct trx_ul_burst_ind *bi)
 {
-	struct l1sched_ts *l1ts = l1sched_trx_get_ts(l1t, bi->tn);
-	struct l1sched_chan_state *chan_state = &l1ts->chan_state[chan];
+	struct l1sched_chan_state *chan_state = &l1ts->chan_state[bi->chan];
 	struct gsm_lchan *lchan = chan_state->lchan;
 	sbit_t *burst, **bursts_p = &chan_state->ul_bursts;
 	uint8_t *mask = &chan_state->ul_mask;
@@ -66,14 +65,14 @@ int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 	uint16_t ber10k;
 	uint8_t is_sub = 0;
 	uint8_t ft;
+	bool amr_is_cmr;
 
 	/* If handover RACH detection is turned on, treat this burst as an Access Burst.
 	 * Handle NOPE.ind as usually to ensure proper Uplink measurement reporting. */
 	if (chan_state->ho_rach_detect == 1 && ~bi->flags & TRX_BI_F_NOPE_IND)
-		return rx_rach_fn(l1t, chan, bid, bi);
+		return rx_rach_fn(l1ts, bi);
 
-	LOGL1S(DL1P, LOGL_DEBUG, l1t, bi->tn, chan, bi->fn,
-	       "Received TCH/F, bid=%u\n", bid);
+	LOGL1SB(DL1P, LOGL_DEBUG, l1ts, bi, "Received TCH/F, bid=%u\n", bi->bid);
 
 	/* allocate burst memory, if not already */
 	if (!*bursts_p) {
@@ -83,19 +82,19 @@ int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 	}
 
 	/* clear burst */
-	if (bid == 0) {
+	if (bi->bid == 0) {
 		memset(*bursts_p + 464, 0, 464);
 		*mask = 0x0;
 	}
 
 	/* update mask */
-	*mask |= (1 << bid);
+	*mask |= (1 << bi->bid);
 
 	/* store measurements */
 	trx_sched_meas_push(chan_state, bi);
 
 	/* copy burst to end of buffer of 8 bursts */
-	burst = *bursts_p + bid * 116 + 464;
+	burst = *bursts_p + bi->bid * 116 + 464;
 	if (bi->burst_len > 0) {
 		memcpy(burst, bi->burst + 3, 58);
 		memcpy(burst + 58, bi->burst + 87, 58);
@@ -103,12 +102,12 @@ int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 		memset(burst, 0, 116);
 
 	/* wait until complete set of bursts */
-	if (bid != 3)
+	if (bi->bid != 3)
 		return 0;
 
 	/* check for complete set of bursts */
 	if ((*mask & 0xf) != 0xf) {
-		LOGL1S(DL1P, LOGL_NOTICE, l1t, bi->tn, chan, bi->fn,
+		LOGL1SB(DL1P, LOGL_NOTICE, l1ts, bi,
 			"Received incomplete frame (%u/%u)\n",
 			bi->fn % l1ts->mf_period, l1ts->mf_period);
 	}
@@ -131,6 +130,8 @@ int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 		 * the first FN 4,13,21 defines that CMR is included in frame.
 		 * NOTE: A frame ends 7 FN after start.
 		 */
+		fn_begin = gsm0502_fn_remap(bi->fn, FN_REMAP_TCH_F);
+		amr_is_cmr = !ul_amr_fn_is_cmi(fn_begin);
 
 		/* The AFS_ONSET frame itself does not result into an RTP frame
 		 * since it only contains a recognition pattern that marks the
@@ -146,15 +147,13 @@ int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 		 * know this before we actually decode the frame) */
 		amr = 2;
 		rc = gsm0503_tch_afs_decode_dtx(tch_data + amr, *bursts_p,
-			(((bi->fn + 26 - 7) % 26) >> 2) & 1, chan_state->codec,
-			chan_state->codecs, &chan_state->ul_ft,
+			amr_is_cmr, chan_state->codec, chan_state->codecs, &chan_state->ul_ft,
 			&chan_state->ul_cmr, &n_errors, &n_bits_total, &chan_state->amr_last_dtx);
 
 		/* Tag all frames that are not regular AMR voice frames as
 		 * SUB-Frames */
 		if (chan_state->amr_last_dtx != AMR_OTHER) {
-			LOGL1S(DL1P, LOGL_DEBUG, l1t, bi->tn, chan, bi->fn,
-			       "Received AMR SID frame: %s\n",
+			LOGL1SB(DL1P, LOGL_DEBUG, l1ts, bi, "Received AMR SID frame: %s\n",
 			       gsm0503_amr_dtx_frame_name(chan_state->amr_last_dtx));
 			is_sub = 1;
 		}
@@ -185,13 +184,11 @@ int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 		}
 
 		if (rc)
-			trx_loop_amr_input(l1t,
-				trx_chan_desc[chan].chan_nr | bi->tn, chan_state,
-				n_errors, n_bits_total);
+			trx_loop_amr_input(chan_state, n_errors, n_bits_total);
 		/* only good speech frames get rtp header */
 		if (rc != GSM_MACBLOCK_LEN && rc >= 4) {
 			if (chan_state->amr_last_dtx == AMR_OTHER) {
-				ft = chan_state->codec[chan_state->ul_cmr];
+				ft = chan_state->codec[chan_state->ul_ft];
 			} else {
 				/* SID frames will always get Frame Type Index 8 (AMR_SID) */
 				ft = AMR_SID;
@@ -203,7 +200,7 @@ int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 
 		break;
 	default:
-		LOGL1S(DL1P, LOGL_ERROR, l1t, bi->tn, chan, bi->fn,
+		LOGL1SB(DL1P, LOGL_ERROR, l1ts, bi,
 			"TCH mode %u invalid, please fix!\n",
 			tch_mode);
 		return -EINVAL;
@@ -215,12 +212,11 @@ int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 
 	/* Check if the frame is bad */
 	if (rc < 0) {
-		LOGL1S(DL1P, LOGL_NOTICE, l1t, bi->tn, chan, bi->fn,
-			"Received bad data (%u/%u)\n",
+		LOGL1SB(DL1P, LOGL_NOTICE, l1ts, bi, "Received bad data (%u/%u)\n",
 			bi->fn % l1ts->mf_period, l1ts->mf_period);
 		bfi_flag = true;
 	} else if (rc < 4) {
-		LOGL1S(DL1P, LOGL_NOTICE, l1t, bi->tn, chan, bi->fn,
+		LOGL1SB(DL1P, LOGL_NOTICE, l1ts, bi,
 			"Received bad data (%u/%u) with invalid codec mode %d\n",
 			bi->fn % l1ts->mf_period, l1ts->mf_period, rc);
 		bfi_flag = true;
@@ -236,7 +232,7 @@ int rx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 	/* FACCH */
 	if (rc == GSM_MACBLOCK_LEN) {
 		fn_begin = gsm0502_fn_remap(bi->fn, FN_REMAP_FACCH_F);
-		_sched_compose_ph_data_ind(l1t, bi->tn, fn_begin, chan,
+		_sched_compose_ph_data_ind(l1ts, fn_begin, bi->chan,
 			tch_data + amr, GSM_MACBLOCK_LEN,
 			meas_avg.rssi, meas_avg.toa256,
 			meas_avg.ci_cb, ber10k,
@@ -280,11 +276,11 @@ bfi:
 				break;
 			case GSM48_CMODE_SPEECH_AMR: /* AMR */
 				rc = osmo_amr_rtp_enc(tch_data,
-					chan_state->codec[chan_state->dl_cmr],
-					chan_state->codec[chan_state->dl_ft],
+					chan_state->codec[chan_state->ul_cmr],
+					chan_state->codec[chan_state->ul_ft],
 					AMR_BAD);
 				if (rc < 2) {
-					LOGL1S(DL1P, LOGL_ERROR, l1t, bi->tn, chan, bi->fn,
+					LOGL1SB(DL1P, LOGL_ERROR, l1ts, bi,
 					       "Failed to encode AMR_BAD frame (rc=%d), "
 					       "not sending BFI\n", rc);
 					return -EINVAL;
@@ -292,7 +288,7 @@ bfi:
 				memset(tch_data + 2, 0, rc - 2);
 				break;
 			default:
-				LOGL1S(DL1P, LOGL_ERROR, l1t, bi->tn, chan, bi->fn,
+				LOGL1SB(DL1P, LOGL_ERROR, l1ts, bi,
 					"TCH mode %u invalid, please fix!\n", tch_mode);
 				return -EINVAL;
 			}
@@ -305,21 +301,21 @@ bfi:
 	/* TCH or BFI */
 compose_l1sap:
 	fn_begin = gsm0502_fn_remap(bi->fn, FN_REMAP_TCH_F);
-	return _sched_compose_tch_ind(l1t, bi->tn, fn_begin, chan, tch_data, rc,
+	return _sched_compose_tch_ind(l1ts, fn_begin, bi->chan, tch_data, rc,
 				      /* FIXME: what should we use for BFI here? */
 				      bfi_flag ? bi->toa256 : meas_avg.toa256, ber10k,
-				      bfi_flag ? bi->rssi : meas_avg.rssi, is_sub);
+				      bfi_flag ? bi->rssi : meas_avg.rssi,
+				      bfi_flag ? bi->ci_cb : meas_avg.ci_cb,
+				      is_sub);
 }
 
 /* common section for generation of TCH bursts (TCH/H and TCH/F).
  * FIXME: this function is over-complicated, refactor / get rid of it. */
-void tx_tch_common(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
-		   enum trx_chan_type chan, uint8_t bid,
+void tx_tch_common(struct l1sched_ts *l1ts, struct trx_dl_burst_req *br,
 		   struct msgb **_msg_tch, struct msgb **_msg_facch)
 {
-	struct l1sched_ts *l1ts = l1sched_trx_get_ts(l1t, tn);
 	struct msgb *msg1, *msg2, *msg_tch = NULL, *msg_facch = NULL;
-	struct l1sched_chan_state *chan_state = &l1ts->chan_state[chan];
+	struct l1sched_chan_state *chan_state = &l1ts->chan_state[br->chan];
 	uint8_t rsl_cmode = chan_state->rsl_cmode;
 	uint8_t tch_mode = chan_state->tch_mode;
 	struct osmo_phsap_prim *l1sap;
@@ -330,13 +326,12 @@ void tx_tch_common(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
 		uint8_t tch_data[GSM_FR_BYTES];
 		int len;
 
-		LOGL1S(DL1P, LOGL_NOTICE, l1t, tn, chan, fn,
-			"Missing TCH bursts detected, sending BFI\n");
+		LOGL1SB(DL1P, LOGL_NOTICE, l1ts, br, "Missing TCH bursts detected, sending BFI\n");
 
 		/* indicate bad frame */
 		switch (tch_mode) {
 		case GSM48_CMODE_SPEECH_V1: /* FR / HR */
-			if (chan != TRXC_TCHF) { /* HR */
+			if (br->chan != TRXC_TCHF) { /* HR */
 				tch_data[0] = 0x70; /* F = 0, FT = 111 */
 				memset(tch_data + 1, 0, 14);
 				len = 15;
@@ -346,7 +341,7 @@ void tx_tch_common(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
 			len = GSM_FR_BYTES;
 			break;
 		case GSM48_CMODE_SPEECH_EFR: /* EFR */
-			if (chan != TRXC_TCHF)
+			if (br->chan != TRXC_TCHF)
 				goto inval_mode1;
 			memset(tch_data, 0, GSM_EFR_BYTES);
 			len = GSM_EFR_BYTES;
@@ -356,29 +351,30 @@ void tx_tch_common(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
 				chan_state->codec[chan_state->dl_cmr],
 				chan_state->codec[chan_state->dl_ft], AMR_BAD);
 			if (len < 2) {
-				LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn,
-				       "Failed to encode AMR_BAD frame (rc=%d), "
-				       "not sending BFI\n", len);
+				LOGL1SB(DL1P, LOGL_ERROR, l1ts, br,
+					"Failed to encode AMR_BAD frame (rc=%d), "
+					"not sending BFI\n", len);
 				return;
 			}
 			memset(tch_data + 2, 0, len - 2);
 			break;
 		default:
 inval_mode1:
-			LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn, "TCH mode invalid, please fix!\n");
+			LOGL1SB(DL1P, LOGL_ERROR, l1ts, br, "TCH mode invalid, please fix!\n");
 			len = 0;
 		}
 
 		if (len) {
 			/* Note: RSSI/ToA256 is set to 0 to indicate to the higher
 			 * layers that this is a faked tch_ind */
-			_sched_compose_tch_ind(l1t, tn, fn, chan, tch_data, len, 0, 10000, 0, 0);
+			_sched_compose_tch_ind(l1ts, br->fn, br->chan,
+					       tch_data, len, 0, 10000, 0, 0, 0);
 		}
 	}
 
 	/* get frame and unlink from queue */
-	msg1 = _sched_dequeue_prim(l1t, tn, fn, chan);
-	msg2 = _sched_dequeue_prim(l1t, tn, fn, chan);
+	msg1 = _sched_dequeue_prim(l1ts, br);
+	msg2 = _sched_dequeue_prim(l1ts, br);
 	if (msg1) {
 		l1sap = msgb_l1sap_prim(msg1);
 		if (l1sap->oph.primitive == PRIM_TCH) {
@@ -386,8 +382,7 @@ inval_mode1:
 			if (msg2) {
 				l1sap = msgb_l1sap_prim(msg2);
 				if (l1sap->oph.primitive == PRIM_TCH) {
-					LOGL1S(DL1P, LOGL_FATAL, l1t, tn, chan, fn,
-						"TCH twice, please FIX!\n");
+					LOGL1SB(DL1P, LOGL_FATAL, l1ts, br, "TCH twice, please FIX!\n");
 					msgb_free(msg2);
 				} else
 					msg_facch = msg2;
@@ -397,25 +392,18 @@ inval_mode1:
 			if (msg2) {
 				l1sap = msgb_l1sap_prim(msg2);
 				if (l1sap->oph.primitive != PRIM_TCH) {
-					LOGL1S(DL1P, LOGL_FATAL, l1t, tn, chan, fn,
-						"FACCH twice, please FIX!\n");
+					LOGL1SB(DL1P, LOGL_FATAL, l1ts, br, "FACCH twice, please FIX!\n");
 					msgb_free(msg2);
 				} else
 					msg_tch = msg2;
 			}
 		}
-	} else if (msg2) {
-		l1sap = msgb_l1sap_prim(msg2);
-		if (l1sap->oph.primitive == PRIM_TCH)
-			msg_tch = msg2;
-		else
-			msg_facch = msg2;
 	}
 
 	/* check validity of message */
 	if (msg_facch && msgb_l2len(msg_facch) != GSM_MACBLOCK_LEN) {
-		LOGL1S(DL1P, LOGL_FATAL, l1t, tn, chan, fn, "Prim not 23 bytes, please FIX! "
-			"(len=%d)\n", msgb_l2len(msg_facch));
+		LOGL1SB(DL1P, LOGL_FATAL, l1ts, br, "Prim has odd len=%u != %u\n",
+			msgb_l2len(msg_facch), GSM_MACBLOCK_LEN);
 		/* free message */
 		msgb_free(msg_facch);
 		msg_facch = NULL;
@@ -429,22 +417,23 @@ inval_mode1:
 		enum osmo_amr_type ft_codec;
 		enum osmo_amr_quality bfi;
 		int8_t sti, cmi;
+		bool amr_is_cmr = !dl_amr_fn_is_cmi(br->fn);
 
 		if (rsl_cmode != RSL_CMOD_SPD_SPEECH) {
-			LOGL1S(DL1P, LOGL_NOTICE, l1t, tn, chan, fn, "Dropping speech frame, "
+			LOGL1SB(DL1P, LOGL_NOTICE, l1ts, br, "Dropping speech frame, "
 				"because we are not in speech mode\n");
 			goto free_bad_msg;
 		}
 
 		switch (tch_mode) {
 		case GSM48_CMODE_SPEECH_V1: /* FR / HR */
-			if (chan != TRXC_TCHF) /* HR */
+			if (br->chan != TRXC_TCHF) /* HR */
 				len = 15;
 			else
 				len = GSM_FR_BYTES;
 			break;
 		case GSM48_CMODE_SPEECH_EFR: /* EFR */
-			if (chan != TRXC_TCHF)
+			if (br->chan != TRXC_TCHF)
 				goto inval_mode2;
 			len = GSM_EFR_BYTES;
 			break;
@@ -469,33 +458,32 @@ inval_mode1:
 				trx_loop_amr_set(chan_state, 1);
 			}
 			if (ft < 0) {
-				LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn,
+				LOGL1SB(DL1P, LOGL_ERROR, l1ts, br,
 					"Codec (FT = %d) of RTP frame not in list\n", ft_codec);
 				goto free_bad_msg;
 			}
-			if (fn_is_codec_mode_request(fn) && chan_state->dl_ft != ft) {
-				LOGL1S(DL1P, LOGL_NOTICE, l1t, tn, chan, fn, "Codec (FT = %d) "
+			if (amr_is_cmr && chan_state->dl_ft != ft) {
+				LOGL1SB(DL1P, LOGL_NOTICE, l1ts, br, "Codec (FT = %d) "
 					" of RTP cannot be changed now, but in next frame\n", ft_codec);
 				goto free_bad_msg;
 			}
 			chan_state->dl_ft = ft;
 			if (bfi == AMR_BAD) {
-				LOGL1S(DL1P, LOGL_NOTICE, l1t, tn, chan, fn,
-					"Transmitting 'bad AMR frame'\n");
+				LOGL1SB(DL1P, LOGL_NOTICE, l1ts, br, "Transmitting 'bad AMR frame'\n");
 				goto free_bad_msg;
 			}
 			break;
 		default:
 inval_mode2:
-			LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn, "TCH mode invalid, please fix!\n");
+			LOGL1SB(DL1P, LOGL_ERROR, l1ts, br, "TCH mode invalid, please fix!\n");
 			goto free_bad_msg;
 		}
 		if (len < 0) {
-			LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn, "Cannot send invalid AMR payload\n");
+			LOGL1SB(DL1P, LOGL_ERROR, l1ts, br, "Cannot send invalid AMR payload\n");
 			goto free_bad_msg;
 		}
 		if (msgb_l2len(msg_tch) != len) {
-			LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn, "Cannot send payload with "
+			LOGL1SB(DL1P, LOGL_ERROR, l1ts, br, "Cannot send payload with "
 				"invalid length! (expecting %d, received %d)\n",
 				len, msgb_l2len(msg_tch));
 free_bad_msg:
@@ -512,24 +500,21 @@ send_frame:
 }
 
 /* obtain a to-be-transmitted TCH/F (Full Traffic Channel) burst */
-int tx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
-	       uint8_t bid, struct trx_dl_burst_req *br)
+int tx_tchf_fn(struct l1sched_ts *l1ts, struct trx_dl_burst_req *br)
 {
 	struct msgb *msg_tch = NULL, *msg_facch = NULL;
-	struct l1sched_ts *l1ts = l1sched_trx_get_ts(l1t, br->tn);
-	struct gsm_bts_trx_ts *ts = &l1t->trx->ts[br->tn];
-	struct l1sched_chan_state *chan_state = &l1ts->chan_state[chan];
+	struct l1sched_chan_state *chan_state = &l1ts->chan_state[br->chan];
 	uint8_t tch_mode = chan_state->tch_mode;
 	ubit_t *burst, **bursts_p = &chan_state->dl_bursts;
 
 	/* send burst, if we already got a frame */
-	if (bid > 0) {
+	if (br->bid > 0) {
 		if (!*bursts_p)
-			return 0;
+			return -ENODEV;
 		goto send_burst;
 	}
 
-	tx_tch_common(l1t, br->tn, br->fn, chan, bid, &msg_tch, &msg_facch);
+	tx_tch_common(l1ts, br, &msg_tch, &msg_facch);
 
 	/* BURST BYPASS */
 
@@ -544,22 +529,30 @@ int tx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 		memset(*bursts_p + 464, 0, 464);
 	}
 
-	/* no message at all */
+	/* no message at all, send a dummy L2 frame on FACCH */
 	if (!msg_tch && !msg_facch) {
-		LOGL1S(DL1P, LOGL_INFO, l1t, br->tn, chan, br->fn, "No TCH or FACCH prim for transmit.\n");
+		static const uint8_t dummy[GSM_MACBLOCK_LEN] = {
+			0x03, 0x03, 0x01, /* TODO: use randomized padding */
+			0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b,
+			0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b,
+		};
+
+		LOGL1SB(DL1P, LOGL_DEBUG, l1ts, br, "No TCH or FACCH prim for transmit.\n");
+		gsm0503_tch_fr_encode(*bursts_p, dummy, sizeof(dummy), 1);
 		goto send_burst;
 	}
 
 	/* encode bursts (prioritize FACCH) */
-	if (msg_facch)
+	if (msg_facch) {
 		gsm0503_tch_fr_encode(*bursts_p, msg_facch->l2h, msgb_l2len(msg_facch),
 			1);
-	else if (tch_mode == GSM48_CMODE_SPEECH_AMR)
+		chan_state->dl_facch_bursts = 8;
+	} else if (tch_mode == GSM48_CMODE_SPEECH_AMR)
 		/* the first FN 4,13,21 defines that CMI is included in frame,
 		 * the first FN 0,8,17 defines that CMR is included in frame.
 		 */
 		gsm0503_tch_afs_encode(*bursts_p, msg_tch->l2h + 2,
-			msgb_l2len(msg_tch) - 2, fn_is_codec_mode_request(br->fn),
+			msgb_l2len(msg_tch) - 2, !dl_amr_fn_is_cmi(br->fn),
 			chan_state->codec, chan_state->codecs,
 			chan_state->dl_ft,
 			chan_state->dl_cmr);
@@ -574,14 +567,19 @@ int tx_tchf_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
 
 send_burst:
 	/* compose burst */
-	burst = *bursts_p + bid * 116;
+	burst = *bursts_p + br->bid * 116;
 	memcpy(br->burst + 3, burst, 58);
-	memcpy(br->burst + 61, _sched_tsc[gsm_ts_tsc(ts)], 26);
+	memcpy(br->burst + 61, TRX_GMSK_NB_TSC(br), 26);
 	memcpy(br->burst + 87, burst + 58, 58);
 
 	br->burst_len = GSM_BURST_LEN;
 
-	LOGL1S(DL1P, LOGL_DEBUG, l1t, br->tn, chan, br->fn, "Transmitting burst=%u.\n", bid);
+	if (chan_state->dl_facch_bursts > 0) {
+		chan_state->dl_facch_bursts--;
+		br->flags |= TRX_BR_F_FACCH;
+	}
+
+	LOGL1SB(DL1P, LOGL_DEBUG, l1ts, br, "Transmitting burst=%u.\n", br->bid);
 
 	return 0;
 }
