@@ -3,6 +3,7 @@
 /* (C) 2013 by Andreas Eversberg <jolly@eversberg.eu>
  * (C) 2015 by Alexander Chemeris <Alexander.Chemeris@fairwaves.co>
  * (C) 2015-2017 by Harald Welte <laforge@gnumonks.org>
+ * (C) 2020-2021 by sysmocom - s.m.f.c. GmbH <info@sysmocom.de>
  *
  * All Rights Reserved
  *
@@ -44,20 +45,92 @@
 #include <osmo-bts/l1sap.h>
 #include <osmo-bts/scheduler.h>
 #include <osmo-bts/scheduler_backend.h>
+#include <osmo-bts/pcu_if.h>
 
 #include "l1_if.h"
 #include "trx_if.h"
 
+#include "btsconfig.h"
+
+#ifdef HAVE_SYSTEMTAP
+/* include the generated probes header and put markers in code */
+#include "probes.h"
+#define TRACE(probe) probe
+#define TRACE_ENABLED(probe) probe ## _ENABLED()
+#else
+/* Wrap the probe to allow it to be removed when no systemtap available */
+#define TRACE(probe)
+#define TRACE_ENABLED(probe) (0)
+#endif /* HAVE_SYSTEMTAP */
+
 #define SCHED_FH_PARAMS_FMT "hsn=%u, maio=%u, ma_len=%u"
 #define SCHED_FH_PARAMS_VALS(ts) \
 	(ts)->hopping.hsn, (ts)->hopping.maio, (ts)->hopping.arfcn_num
 
-/* an IDLE burst returns nothing. on C0 it is replaced by dummy burst */
-int tx_idle_fn(struct l1sched_trx *l1t, enum trx_chan_type chan,
-	       uint8_t bid, struct trx_dl_burst_req *br)
+static void lchan_report_interf_meas(const struct gsm_lchan *lchan)
 {
-	LOGL1S(DL1P, LOGL_DEBUG, l1t, br->tn, chan, br->fn, "Transmitting IDLE\n");
-	return 0;
+	const struct gsm_bts_trx_ts *ts = lchan->ts;
+	const struct l1sched_ts *l1ts = ts->priv;
+	enum trx_chan_type dcch, acch;
+	int interf_avg;
+
+	/* We're not interested in active CS channels */
+	if (lchan->state == LCHAN_S_ACTIVE) {
+		if (lchan->type != GSM_LCHAN_PDTCH)
+			return;
+	}
+
+	switch (lchan->type) {
+	case GSM_LCHAN_SDCCH:
+		if (ts->pchan == GSM_PCHAN_CCCH_SDCCH4 ||
+		    ts->pchan == GSM_PCHAN_CCCH_SDCCH4_CBCH) {
+			dcch = TRXC_SDCCH4_0 + lchan->nr;
+			acch = TRXC_SACCH4_0 + lchan->nr;
+		} else { /* SDCCH/8 otherwise */
+			dcch = TRXC_SDCCH8_0 + lchan->nr;
+			acch = TRXC_SACCH8_0 + lchan->nr;
+		}
+		break;
+	case GSM_LCHAN_TCH_F:
+		dcch = TRXC_TCHF;
+		acch = TRXC_SACCHTF;
+		break;
+	case GSM_LCHAN_TCH_H:
+		dcch = TRXC_TCHH_0 + lchan->nr;
+		acch = TRXC_SACCHTH_0 + lchan->nr;
+		break;
+	case GSM_LCHAN_PDTCH:
+		/* We use idle TDMA frames on PDCH */
+		dcch = TRXC_IDLE;
+		acch = TRXC_IDLE;
+		break;
+	default:
+		/* Skip other lchan types */
+		return;
+	}
+
+	OSMO_ASSERT(dcch < ARRAY_SIZE(l1ts->chan_state));
+	OSMO_ASSERT(acch < ARRAY_SIZE(l1ts->chan_state));
+
+	interf_avg = (l1ts->chan_state[dcch].meas.interf_avg +
+		      l1ts->chan_state[acch].meas.interf_avg) / 2;
+
+	gsm_lchan_interf_meas_push((struct gsm_lchan *) lchan, interf_avg);
+}
+
+static void bts_report_interf_meas(const struct gsm_bts *bts,
+				   const uint32_t fn)
+{
+	const struct gsm_bts_trx *trx;
+	unsigned int tn, ln;
+
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+			const struct gsm_bts_trx_ts *ts = &trx->ts[tn];
+			for (ln = 0; ln < ARRAY_SIZE(ts->lchan); ln++)
+				lchan_report_interf_meas(&ts->lchan[ln]);
+		}
+	}
 }
 
 /* Find a route (PHY instance) for a given Downlink burst request */
@@ -73,13 +146,13 @@ static struct phy_instance *dlfh_route_br(const struct trx_dl_burst_req *br,
 	/* Check the "cache" first, so we eliminate frequent lookups */
 	idx = gsm0502_hop_seq_gen(&time, SCHED_FH_PARAMS_VALS(ts), NULL);
 	if (ts->fh_trx_list[idx] != NULL)
-		return ts->fh_trx_list[idx]->role_bts.l1h;
+		return ts->fh_trx_list[idx]->pinst;
 
 	/* The "cache" may not be filled yet, lookup the transceiver */
 	llist_for_each_entry(trx, &ts->trx->bts->trx_list, list) {
 		if (trx->arfcn == ts->hopping.arfcn_list[idx]) {
 			ts->fh_trx_list[idx] = trx;
-			return trx->role_bts.l1h;
+			return trx->pinst;
 		}
 	}
 
@@ -88,93 +161,165 @@ static struct phy_instance *dlfh_route_br(const struct trx_dl_burst_req *br,
 		br->fn, br->tn, SCHED_FH_PARAMS_VALS(ts));
 
 	struct bts_trx_priv *priv = (struct bts_trx_priv *) ts->trx->bts->model_priv;
-	rate_ctr_inc(&priv->ctrs->ctr[BTSTRX_CTR_SCHED_DL_FH_NO_CARRIER]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(priv->ctrs, BTSTRX_CTR_SCHED_DL_FH_NO_CARRIER));
 
 	return NULL;
 }
 
-/* schedule all frames of all TRX for given FN */
-static void trx_sched_fn(struct gsm_bts *bts, const uint32_t fn)
+static void bts_sched_init_buffers(struct gsm_bts *bts, const uint32_t fn)
 {
-	struct trx_dl_burst_req br;
 	struct gsm_bts_trx *trx;
-	uint8_t c0_mask = 0x00;
-	uint32_t sched_fn;
 	uint8_t tn;
 
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		struct phy_instance *pinst = trx->pinst;
+		const struct phy_link *plink = pinst->phy_link;
+
+		/* Advance frame number, so the PHY has more time to process bursts */
+		const uint32_t sched_fn = GSM_TDMA_FN_SUM(fn, plink->u.osmotrx.clock_advance);
+
+		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+			struct trx_dl_burst_req *br = &pinst->u.osmotrx.br[tn];
+
+			*br = (struct trx_dl_burst_req) {
+				.trx_num = trx->nr,
+				.fn = sched_fn,
+				.tn = tn,
+			};
+		}
+	}
+
+	/* Initialize all timeslots on C0/TRX0 with dummy burst */
+	for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+		struct phy_instance *pinst = bts->c0->pinst;
+		struct trx_dl_burst_req *br = &pinst->u.osmotrx.br[tn];
+		const struct gsm_bts_trx_ts *ts = &bts->c0->ts[tn];
+
+		memcpy(br->burst, _sched_dummy_burst, GSM_BURST_LEN);
+		br->burst_len = GSM_BURST_LEN;
+
+		/* BCCH carrier power reduction for this timeslot */
+		br->att = ts->c0_power_red_db;
+	}
+}
+
+static void bts_sched_flush_buffers(struct gsm_bts *bts)
+{
+	const struct gsm_bts_trx *trx;
+	unsigned int tn;
+
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		const struct phy_instance *pinst = trx->pinst;
+		struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
+
+		for (tn = 0; tn < TRX_NR_TS; tn++) {
+			const struct trx_dl_burst_req *br;
+
+			br = &pinst->u.osmotrx.br[tn];
+			if (!br->burst_len)
+				continue;
+			trx_if_send_burst(l1h, br);
+		}
+
+		/* Batch all timeslots into a single TRXD PDU */
+		trx_if_send_burst(l1h, NULL);
+	}
+}
+
+/* schedule one frame for a shadow timeslot, merge bursts */
+static void _sched_dl_shadow_burst(const struct gsm_bts_trx_ts *ts,
+				   struct trx_dl_burst_req *br)
+{
+	struct l1sched_ts *l1ts = ts->priv;
+
+	/* For the shadow timeslots, physical channel type can be either
+	 * GSM_PCHAN_TCH_{F,H} or GSM_PCHAN_NONE.  Even if the primary
+	 * timeslot is a dynamic timeslot, it's always a concrete value. */
+	if (ts->pchan == GSM_PCHAN_NONE)
+		return;
+
+	struct trx_dl_burst_req sbr = {
+		.trx_num = br->trx_num,
+		.fn = br->fn,
+		.tn = br->tn,
+	};
+
+	_sched_dl_burst(l1ts, &sbr);
+
+	if (br->burst_len != 0 && sbr.burst_len != 0) { /* Both present */
+		memcpy(br->burst + GSM_BURST_LEN, sbr.burst, GSM_BURST_LEN);
+		br->burst_len = 2 * GSM_BURST_LEN;
+		br->mod = TRX_MOD_T_AQPSK;
+		/* FIXME: SCPIR is hard-coded to 0 */
+	} else if (br->burst_len == 0) {
+		/* No primary burst, send shadow burst alone */
+		memcpy(br, &sbr, sizeof(sbr));
+	} else if (sbr.burst_len == 0) {
+		/* No shadow burst, send primary burst alone */
+		return;
+	}
+}
+
+/* schedule all frames of all TRX for given FN */
+static void bts_sched_fn(struct gsm_bts *bts, const uint32_t fn)
+{
+	struct gsm_bts_trx *trx;
+	unsigned int tn;
+
+	/* Report interference measurements */
+	if (fn % 104 == 0) /* SACCH period */
+		bts_report_interf_meas(bts, fn);
+
 	/* send time indication */
 	l1if_mph_time_ind(bts, fn);
 
-	/* process every TRX */
+	/* Initialize Downlink burst buffers */
+	bts_sched_init_buffers(bts, fn);
+
+	/* Populate Downlink burst buffers for each TRX/TS */
 	llist_for_each_entry(trx, &bts->trx_list, list) {
-		const struct phy_instance *_pinst = trx_phy_instance(trx);
-		const struct phy_link *plink = _pinst->phy_link;
-		struct trx_l1h *l1h = _pinst->u.osmotrx.hdl;
-		struct l1sched_trx *l1t = &l1h->l1s;
+		const struct phy_link *plink = trx->pinst->phy_link;
+		struct trx_l1h *l1h = trx->pinst->u.osmotrx.hdl;
 
 		/* we don't schedule, if power is off */
 		if (!trx_if_powered(l1h))
 			continue;
 
-		/* advance frame number, so the transceiver has more
-		 * time until it must be transmitted. */
-		sched_fn = GSM_TDMA_FN_SUM(fn, plink->u.osmotrx.clock_advance);
-
 		/* process every TS of TRX */
-		for (tn = 0; tn < ARRAY_SIZE(l1t->ts); tn++) {
-			const struct phy_instance *pinst = _pinst;
+		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+			struct phy_instance *pinst = trx->pinst;
+			struct gsm_bts_trx_ts *ts = &trx->ts[tn];
+			struct l1sched_ts *l1ts = ts->priv;
+			struct trx_dl_burst_req *br;
 
 			/* ready-to-send */
-			_sched_rts(l1t, tn, GSM_TDMA_FN_SUM(sched_fn, plink->u.osmotrx.rts_advance));
+			TRACE(OSMO_BTS_TRX_DL_RTS_START(trx->nr, tn, fn));
+			_sched_rts(l1ts, GSM_TDMA_FN_SUM(fn, plink->u.osmotrx.clock_advance
+							   + plink->u.osmotrx.rts_advance));
+			TRACE(OSMO_BTS_TRX_DL_RTS_DONE(trx->nr, tn, fn));
 
-			/* All other parameters to be set by _sched_dl_burst() */
-			br = (struct trx_dl_burst_req) {
-				.fn = sched_fn,
-				.tn = tn,
-			};
-
-			/* get burst for FN */
-			_sched_dl_burst(l1t, &br);
-			if (br.burst_len == 0) {
-				/* if no bits, send no burst */
-				continue;
-			}
+			/* pre-initialized buffer for the Downlink burst */
+			br = &pinst->u.osmotrx.br[tn];
 
 			/* resolve PHY instance if freq. hopping is enabled */
-			if (trx->ts[tn].hopping.enabled) {
-				pinst = dlfh_route_br(&br, &trx->ts[tn]);
+			if (ts->hopping.enabled) {
+				pinst = dlfh_route_br(br, ts);
 				if (pinst == NULL)
 					continue;
+				/* simply use a different buffer */
+				br = &pinst->u.osmotrx.br[tn];
 			}
 
-			if (pinst->trx == bts->c0) {
-				/* update dummy burst mask for C0 */
-				c0_mask |= (1 << tn);
-				/* ensure no attenuation on C0 */
-				br.att = 0;
-			}
+			/* get burst for the primary timeslot */
+			_sched_dl_burst(l1ts, br);
 
-			trx_if_send_burst(pinst->u.osmotrx.hdl, &br);
+			/* get burst for the shadow timeslot */
+			_sched_dl_shadow_burst(ts->vamos.peer, br);
 		}
 	}
 
-	/* send dummy bursts on inactive timeslots of C0 */
-	struct phy_instance *pinst = trx_phy_instance(bts->c0);
-	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
-	struct phy_link *plink = pinst->phy_link;
-
-	br = (struct trx_dl_burst_req) {
-		.fn = GSM_TDMA_FN_SUM(fn, plink->u.osmotrx.clock_advance),
-		.burst_len = GSM_BURST_LEN,
-	};
-
-	memcpy(br.burst, _sched_dummy_burst, GSM_BURST_LEN);
-
-	for (br.tn = 0; br.tn < TRX_NR_TS; br.tn++) {
-		if (c0_mask & (1 << br.tn))
-			continue;
-		trx_if_send_burst(l1h, &br);
-	}
+	/* Send everything to the PHY */
+	bts_sched_flush_buffers(bts);
 }
 
 /* Find a route (TRX instance) for a given Uplink burst indication */
@@ -202,30 +347,23 @@ static struct gsm_bts_trx *ulfh_route_bi(const struct trx_ul_burst_ind *bi,
 		bi->fn, bi->tn, SCHED_FH_PARAMS_VALS(&src_trx->ts[bi->tn]));
 
 	struct bts_trx_priv *priv = (struct bts_trx_priv *) src_trx->bts->model_priv;
-	rate_ctr_inc(&priv->ctrs->ctr[BTSTRX_CTR_SCHED_UL_FH_NO_CARRIER]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(priv->ctrs, BTSTRX_CTR_SCHED_UL_FH_NO_CARRIER));
 
 	return NULL;
 }
 
 /* Route a given Uplink burst indication to the scheduler depending on freq. hopping state */
-int trx_sched_route_burst_ind(struct trx_ul_burst_ind *bi, struct l1sched_trx *l1t)
+int trx_sched_route_burst_ind(const struct gsm_bts_trx *trx, struct trx_ul_burst_ind *bi)
 {
-	const struct phy_instance *pinst;
-	const struct gsm_bts_trx *trx;
-	struct trx_l1h *l1h;
-
 	/* no frequency hopping => nothing to do */
-	if (!l1t->trx->ts[bi->tn].hopping.enabled)
-		return trx_sched_ul_burst(l1t, bi);
+	if (!trx->ts[bi->tn].hopping.enabled)
+		return trx_sched_ul_burst(trx->ts[bi->tn].priv, bi);
 
-	trx = ulfh_route_bi(bi, l1t->trx);
+	trx = ulfh_route_bi(bi, trx);
 	if (trx == NULL)
 		return -ENODEV;
 
-	pinst = trx_phy_instance(trx);
-	l1h = pinst->u.osmotrx.hdl;
-
-	return trx_sched_ul_burst(&l1h->l1s, bi);
+	return trx_sched_ul_burst(trx->ts[bi->tn].priv, bi);
 }
 
 /*! maximum number of 'missed' frame periods we can tolerate of OS doesn't schedule us*/
@@ -281,7 +419,7 @@ static int trx_fn_timer_cb(struct osmo_fd *ofd, unsigned int what)
 	if (expire_count > 1) {
 		LOGP(DL1C, LOGL_NOTICE, "FN timer expire_count=%"PRIu64": We missed %"PRIu64" timers\n",
 		     expire_count, expire_count - 1);
-		rate_ctr_add(&bts_trx->ctrs->ctr[BTSTRX_CTR_SCHED_DL_MISS_FN], expire_count - 1);
+		rate_ctr_add(rate_ctr_group_get_ctr(bts_trx->ctrs, BTSTRX_CTR_SCHED_DL_MISS_FN), expire_count - 1);
 	}
 
 	/* check if transceiver is still alive */
@@ -307,9 +445,9 @@ static int trx_fn_timer_cb(struct osmo_fd *ofd, unsigned int what)
 		goto no_clock;
 	}
 
-	/* call trx_sched_fn() for all expired FN */
+	/* call bts_sched_fn() for all expired FN */
 	for (i = 0; i < expire_count; i++)
-		trx_sched_fn(bts, GSM_TDMA_FN_INC(tcs->last_fn_timer.fn));
+		bts_sched_fn(bts, GSM_TDMA_FN_INC(tcs->last_fn_timer.fn));
 
 	return 0;
 
@@ -370,17 +508,14 @@ int trx_sched_clock_stopped(struct gsm_bts *bts)
 static int trx_setup_clock(struct gsm_bts *bts, struct osmo_trx_clock_state *tcs,
 	struct timespec *tv_now, const struct timespec *interval, uint32_t fn)
 {
-	tcs->last_fn_timer.fn = fn;
-	/* call trx cheduler function for new 'last' FN */
-	trx_sched_fn(bts, tcs->last_fn_timer.fn);
-
 	/* schedule first FN clock timer */
 	osmo_timerfd_setup(&tcs->fn_timer_ofd, trx_fn_timer_cb, bts);
 	osmo_timerfd_schedule(&tcs->fn_timer_ofd, NULL, interval);
 
+	tcs->last_fn_timer.fn = fn;
 	tcs->last_fn_timer.tv = *tv_now;
-	tcs->last_clk_ind.tv = *tv_now;
-	tcs->last_clk_ind.fn = fn;
+	/* call trx scheduler function for new 'last' FN */
+	bts_sched_fn(bts, tcs->last_fn_timer.fn);
 
 	return 0;
 }
@@ -455,7 +590,7 @@ int trx_sched_clock(struct gsm_bts *bts, uint32_t fn)
 
 	/* transmit what we still need to transmit */
 	while (fn != tcs->last_fn_timer.fn) {
-		trx_sched_fn(bts, GSM_TDMA_FN_INC(tcs->last_fn_timer.fn));
+		bts_sched_fn(bts, GSM_TDMA_FN_INC(tcs->last_fn_timer.fn));
 		fn_caught_up++;
 	}
 
@@ -467,9 +602,9 @@ int trx_sched_clock(struct gsm_bts *bts, uint32_t fn)
 	return 0;
 }
 
-void _sched_act_rach_det(struct l1sched_trx *l1t, uint8_t tn, uint8_t ss, int activate)
+void _sched_act_rach_det(struct gsm_bts_trx *trx, uint8_t tn, uint8_t ss, int activate)
 {
-	struct phy_instance *pinst = trx_phy_instance(l1t->trx);
+	struct phy_instance *pinst = trx_phy_instance(trx);
 	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
 
 	if (activate)
