@@ -8,6 +8,7 @@
  * Copyright (C) 2013  Andreas Eversberg <jolly@eversberg.eu>
  * Copyright (C) 2016-2017  Harald Welte <laforge@gnumonks.org>
  * Copyright (C) 2019  Vadim Yanitskiy <axilirator@gmail.com>
+ * Copyright (C) 2021  sysmocom - s.m.f.c. GmbH <info@sysmocom.de>
  *
  * All Rights Reserved
  *
@@ -50,6 +51,19 @@
 #include "trx_if.h"
 #include "trx_provision_fsm.h"
 
+#include "btsconfig.h"
+
+#ifdef HAVE_SYSTEMTAP
+/* include the generated probes header and put markers in code */
+#include "probes.h"
+#define TRACE(probe) probe
+#define TRACE_ENABLED(probe) probe ## _ENABLED()
+#else
+/* Wrap the probe to allow it to be removed when no systemtap available */
+#define TRACE(probe)
+#define TRACE_ENABLED(probe) (0)
+#endif /* HAVE_SYSTEMTAP */
+
 /*
  * socket helper functions
  */
@@ -95,7 +109,7 @@ static int trx_clk_read_cb(struct osmo_fd *ofd, unsigned int what)
 {
 	struct phy_link *plink = ofd->data;
 	struct phy_instance *pinst = phy_instance_by_num(plink, 0);
-	char buf[1500];
+	char buf[TRXC_MSG_BUF_SIZE];
 	ssize_t len;
 	uint32_t fn;
 
@@ -148,7 +162,7 @@ static int trx_clk_read_cb(struct osmo_fd *ofd, unsigned int what)
 static void trx_ctrl_send(struct trx_l1h *l1h)
 {
 	struct trx_ctrl_msg *tcm;
-	char buf[1500];
+	char buf[TRXC_MSG_BUF_SIZE];
 	int len;
 	ssize_t snd_len;
 
@@ -193,6 +207,12 @@ void trx_if_init(struct trx_l1h *l1h)
 {
 	l1h->trx_ctrl_timer.cb = trx_ctrl_timer_cb;
 	l1h->trx_ctrl_timer.data = l1h;
+
+	/* initialize ctrl queue */
+	INIT_LLIST_HEAD(&l1h->trx_ctrl_list);
+
+	l1h->trx_ofd_ctrl.fd = -1;
+	l1h->trx_ofd_data.fd = -1;
 }
 
 /*! Send a new TRX control command.
@@ -269,11 +289,11 @@ int trx_if_cmd_poweron(struct trx_l1h *l1h, trx_if_cmd_poweronoff_cb *cb)
 	return trx_ctrl_cmd_cb(l1h, 1, cb, "POWERON", "");
 }
 
-/*! Send "SETFORMAT" command to TRX: change TRXD header format version */
+/*! Send "SETFORMAT" command to TRX: change TRXD PDU version */
 int trx_if_cmd_setformat(struct trx_l1h *l1h, uint8_t ver, trx_if_cmd_generic_cb *cb)
 {
 	LOGPPHI(l1h->phy_inst, DTRX, LOGL_INFO,
-		"Requesting TRXD header format version %u\n", ver);
+		"Requesting TRXD PDU version %u\n", ver);
 
 	return trx_ctrl_cmd_cb(l1h, 0, cb, "SETFORMAT", "%u", ver);
 }
@@ -320,10 +340,23 @@ int trx_if_cmd_setmaxdlynb(struct trx_l1h *l1h, int dly)
 	return trx_ctrl_cmd(l1h, 0, "SETMAXDLYNB", "%d", dly);
 }
 
-/*! Send "SETSLOT" command to TRX: Configure Channel Combination for TS */
-int trx_if_cmd_setslot(struct trx_l1h *l1h, uint8_t tn, uint8_t type, trx_if_cmd_setslot_cb *cb)
+/*! Send "SETSLOT" command to TRX: Configure Channel Combination and TSC for TS */
+int trx_if_cmd_setslot(struct trx_l1h *l1h, uint8_t tn,
+		       trx_if_cmd_setslot_cb *cb)
 {
-	return trx_ctrl_cmd_cb(l1h, 1, cb, "SETSLOT", "%d %d", tn, type);
+	const struct trx_config *cfg = &l1h->config;
+	const struct phy_instance *pinst = l1h->phy_inst;
+
+	if (cfg->setslot[tn].tsc_valid && cfg->setslot[tn].tsc_val != BTS_TSC(pinst->trx->bts)) {
+		/* PHY is instructed to use a custom TSC */
+		return trx_ctrl_cmd_cb(l1h, 1, cb, "SETSLOT", "%u %u C%u/S%u",
+				       tn, cfg->setslot[tn].slottype,
+				       cfg->setslot[tn].tsc_val,
+				       cfg->setslot[tn].tsc_set);
+       } else { /* PHY is instructed to use the default TSC from 'SETTSC' */
+		return trx_ctrl_cmd_cb(l1h, 1, cb, "SETSLOT", "%u %u",
+				       tn, cfg->setslot[tn].slottype);
+	}
 }
 
 /*! Send "RXTUNE" command to TRX: Tune Receiver to given ARFCN */
@@ -391,6 +424,7 @@ struct trx_ctrl_rsp {
 
 static int parse_rsp(const char *buf_in, size_t len_in, struct trx_ctrl_rsp *rsp)
 {
+	size_t nlen, plen;
 	char *p, *k;
 
 	if (strncmp(buf_in, "RSP ", 4))
@@ -400,14 +434,17 @@ static int parse_rsp(const char *buf_in, size_t len_in, struct trx_ctrl_rsp *rsp
 	if (!(p = strchr(buf_in + 4, ' ')))
 		goto parse_err;
 
-	if (p - buf_in >= sizeof(rsp->cmd)) {
-		LOGP(DTRX, LOGL_ERROR, "cmd buffer too small %lu >= %zu\n",
-		     (long unsigned) (p - buf_in), sizeof(rsp->cmd));
+	/* Calculate length of the name part */
+	nlen = p - (buf_in + 4);
+
+	if (nlen >= sizeof(rsp->cmd)) {
+		LOGP(DTRX, LOGL_ERROR, "TRXC command name part is too long: "
+		     "%zu >= %zu\n", nlen, sizeof(rsp->cmd));
 		goto parse_err;
 	}
 
-	rsp->cmd[0] = '\0';
-	strncat(rsp->cmd, buf_in + 4, p - buf_in - 4);
+	memcpy(&rsp->cmd[0], buf_in + 4, nlen);
+	rsp->cmd[nlen] = '\0';
 
 	/* Now comes the status code of the response */
 	p++;
@@ -421,18 +458,22 @@ static int parse_rsp(const char *buf_in, size_t len_in, struct trx_ctrl_rsp *rsp
 	else
 		k = p + strlen(p);
 
-	if (strlen(k) >= sizeof(rsp->params)) {
-		LOGP(DTRX, LOGL_ERROR, "params buffer too small %zu >= %zu\n",
-			strlen(k), sizeof(rsp->params));
+	/* Calculate length of the parameters part */
+	plen = strlen(k);
+
+	if (plen >= sizeof(rsp->params)) {
+		LOGP(DTRX, LOGL_ERROR, "TRXC command parameters part is too long: "
+		     "%zu >= %zu\n", plen, sizeof(rsp->params));
 		goto parse_err;
 	}
-	rsp->params[0] = '\0';
-	strcat(rsp->params, k);
+
+	memcpy(&rsp->params[0], k, plen);
+	rsp->params[plen] = '\0';
+
 	return 0;
 
 parse_err:
-	LOGP(DTRX, LOGL_NOTICE, "Unknown message on ctrl port: %s\n",
-		buf_in);
+	LOGP(DTRX, LOGL_NOTICE, "Unknown TRXC message: %s\n", buf_in);
 	return -1;
 }
 
@@ -504,7 +545,7 @@ static int trx_ctrl_rx_rsp_setslot(struct trx_l1h *l1h, struct trx_ctrl_rsp *rsp
 	return rsp->status == 0 ? 0 : -EINVAL;
 }
 
-/* TRXD header format negotiation handler.
+/* TRXD PDU format negotiation handler.
  *
  * If the transceiver does not support the format negotiation, it would
  * reject SETFORMAT with 'RSP ERR 1'. If the requested version is not
@@ -520,7 +561,7 @@ static int trx_ctrl_rx_rsp_setformat(struct trx_l1h *l1h,
 	if (strcmp(rsp->cmd, "SETFORMAT") != 0) {
 		LOGPPHI(l1h->phy_inst, DTRX, LOGL_NOTICE,
 			"Transceiver rejected the format negotiation command, "
-			"using legacy TRXD header format version (0)\n");
+			"using legacy TRXD PDU version (0)\n");
 		if (rsp->cb) {
 			cb = (trx_if_cmd_generic_cb*) rsp->cb;
 			cb(l1h, 0);
@@ -529,11 +570,11 @@ static int trx_ctrl_rx_rsp_setformat(struct trx_l1h *l1h,
 	}
 
 	/* Status shall indicate a proper version supported by the transceiver */
-	if (rsp->status < 0 || rsp->status > l1h->config.trxd_hdr_ver_req) {
+	if (rsp->status < 0 || rsp->status > l1h->config.trxd_pdu_ver_req) {
 		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
 			"Transceiver indicated an out of range "
-			"header format version %d (requested %u)\n",
-			rsp->status, l1h->config.trxd_hdr_ver_req);
+			"PDU version %d (requested %u)\n",
+			rsp->status, l1h->config.trxd_pdu_ver_req);
 		return -EINVAL;
 	}
 
@@ -628,7 +669,7 @@ static int trx_ctrl_read_cb(struct osmo_fd *ofd, unsigned int what)
 {
 	struct trx_l1h *l1h = ofd->data;
 	struct phy_instance *pinst = l1h->phy_inst;
-	char buf[1500];
+	char buf[TRXC_MSG_BUF_SIZE];
 	struct trx_ctrl_rsp rsp;
 	int len, rc;
 	struct trx_ctrl_msg *tcm;
@@ -681,6 +722,13 @@ static int trx_ctrl_read_cb(struct osmo_fd *ofd, unsigned int what)
 
 	rsp.cb = tcm->cb;
 
+	/* Remove command from list, save it to last_acked and remove previous
+	 * last_acked. Do it before calling callback to avoid user freeing tcm
+	 * pointer if flushing/closing the iface. */
+	llist_del(&tcm->list);
+	talloc_free(l1h->last_acked);
+	l1h->last_acked = tcm;
+
 	/* check for response code */
 	rc = trx_ctrl_rx_rsp(l1h, &rsp, tcm);
 	if (rc == -EINVAL)
@@ -688,15 +736,11 @@ static int trx_ctrl_read_cb(struct osmo_fd *ofd, unsigned int what)
 
 	/* re-schedule last cmd in rc seconds time */
 	if (rc > 0) {
-		osmo_timer_schedule(&l1h->trx_ctrl_timer, rc, 0);
+		if (!llist_empty(&l1h->trx_ctrl_list))
+			osmo_timer_schedule(&l1h->trx_ctrl_timer, rc, 0);
 		return 0;
 	}
 
-	/* remove command from list, save it to last_acked and removed previous last_acked */
-	llist_del(&tcm->list);
-	talloc_free(l1h->last_acked);
-	l1h->last_acked = tcm;
-
 	trx_ctrl_send(l1h);
 
 	return 0;
@@ -712,95 +756,115 @@ rsp_error:
  * TRX burst data socket
  */
 
-/* Maximum DATA message length (header + burst) */
-#define TRX_DATA_MSG_MAX_LEN	512
-
-/* Common header length: 1/2 VER + 1/2 TDMA TN + 4 TDMA FN */
-#define TRX_CHDR_LEN		(1 + 4)
-/* Uplink v0 header length: 1 RSSI + 2 ToA256 */
-#define TRX_UL_V0HDR_LEN	(TRX_CHDR_LEN + 1 + 2)
-/* Uplink v1 header length: + 1 MTS + 2 C/I */
+/* Uplink TRXDv0 header length: TDMA TN + FN + RSSI + ToA256 */
+#define TRX_UL_V0HDR_LEN	(1 + 4 + 1 + 2)
+/* Uplink TRXDv1 header length: additional MTS + C/I */
 #define TRX_UL_V1HDR_LEN	(TRX_UL_V0HDR_LEN + 1 + 2)
+/* Uplink TRXDv2 header length: TDMA TN + TRXN + MTS + RSSI + ToA256 + C/I */
+#define TRX_UL_V2HDR_LEN	(1 + 1 + 1 + 1 + 2 + 2)
 
-/* TRXD header dissector for version 0 */
-static int trx_data_handle_hdr_v0(struct trx_l1h *l1h,
-				  struct trx_ul_burst_ind *bi,
-				  const uint8_t *buf, size_t buf_len)
+/* Minimum Uplink TRXD header length for all PDU versions */
+static const uint8_t trx_data_rx_hdr_len[] = {
+	TRX_UL_V0HDR_LEN, /* TRXDv0 */
+	TRX_UL_V1HDR_LEN, /* TRXDv1 */
+	TRX_UL_V2HDR_LEN, /* TRXDv2 */
+};
+
+static const uint8_t trx_data_mod_val[] = {
+	[TRX_MOD_T_GMSK]	= 0x00, /* .00xx... */
+	[TRX_MOD_T_8PSK]	= 0x20, /* .010x... */
+	[TRX_MOD_T_AQPSK]	= 0x60, /* .11xx... */
+};
+
+/* Header dissector for TRXDv0 (and part of TRXDv1) */
+static inline void trx_data_handle_hdr_v0_part(struct trx_ul_burst_ind *bi,
+					       const uint8_t *buf)
 {
-	/* Make sure we have enough data */
-	if (buf_len < TRX_UL_V0HDR_LEN) {
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
-			"Short read on TRXD, missing version 0 header "
-			"(len=%zu vs expected %d)\n", buf_len, TRX_UL_V0HDR_LEN);
-		return -EIO;
-	}
-
 	bi->tn = buf[0] & 0b111;
 	bi->fn = osmo_load32be(buf + 1);
 	bi->rssi = -(int8_t)buf[5];
 	bi->toa256 = (int16_t) osmo_load16be(buf + 6);
+}
 
-	if (bi->fn >= GSM_TDMA_HYPERFRAME) {
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
-			"Illegal TDMA fn=%u\n", bi->fn);
+/* TRXD header dissector for version 0x00 */
+static int trx_data_handle_hdr_v0(struct phy_instance *phy_inst,
+				  struct trx_ul_burst_ind *bi,
+				  const uint8_t *buf, size_t buf_len)
+{
+	/* Parse TRXDv0 specific header part */
+	trx_data_handle_hdr_v0_part(bi, buf);
+	buf_len -= TRX_UL_V0HDR_LEN;
+
+	/* Guess modulation and burst length by the rest octets.
+	 * NOTE: a legacy transceiver may append two garbage bytes. */
+	switch (buf_len) {
+	case EGPRS_BURST_LEN + 2:
+	case EGPRS_BURST_LEN:
+		bi->mod = TRX_MOD_T_8PSK;
+		break;
+	case GSM_BURST_LEN + 2:
+	case GSM_BURST_LEN:
+		bi->mod = TRX_MOD_T_GMSK;
+		break;
+	default:
+		LOGPPHI(phy_inst, DTRX, LOGL_NOTICE,
+			"Rx TRXD PDU with odd burst length %zu\n", buf_len);
 		return -EINVAL;
 	}
 
 	return TRX_UL_V0HDR_LEN;
 }
 
-/* TRXD header dissector for version 0x01 */
-static int trx_data_handle_hdr_v1(struct trx_l1h *l1h,
-				  struct trx_ul_burst_ind *bi,
-				  const uint8_t *buf, size_t buf_len)
+/* Parser for MTS (Modulation and Training Sequence) */
+static inline int trx_data_parse_mts(struct phy_instance *phy_inst,
+				     struct trx_ul_burst_ind *bi,
+				     const uint8_t mts)
 {
-	uint8_t mts;
-	int rc;
-
-	/* Make sure we have enough data */
-	if (buf_len < TRX_UL_V1HDR_LEN) {
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
-			"Short read on TRXD, missing version 1 header "
-			"(len=%zu vs expected %d)\n", buf_len, TRX_UL_V1HDR_LEN);
-		return -EIO;
-	}
-
-	/* Parse v0 specific part */
-	rc = trx_data_handle_hdr_v0(l1h, bi, buf, buf_len);
-	if (rc < 0)
-		return rc;
-
-	/* Move closer to the v1 specific part */
-	buf_len -= rc;
-	buf += rc;
-
-	/* IDLE / NOPE frame indication */
-	if (buf[0] & (1 << 7)) {
+	if (mts & (1 << 7)) {
 		bi->flags |= TRX_BI_F_NOPE_IND;
-		goto skip_mts;
+		return 0;
 	}
 
-	/* Modulation info and TSC set */
-	mts = (buf[0] >> 3) & 0b1111;
-	if ((mts & 0b1100) == 0x00) {
-		bi->bt = TRX_BURST_GMSK;
-		bi->tsc_set = mts & 0b11;
-		bi->flags |= TRX_BI_F_MOD_TYPE;
-	} else if ((mts & 0b0100) == 0b0100) {
-		bi->bt = TRX_BURST_8PSK;
-		bi->tsc_set = mts & 0b1;
-		bi->flags |= TRX_BI_F_MOD_TYPE;
+	/* | 7 6 5 4 3 2 1 0 | Bitmask / description
+	 * | . 0 0 X X . . . | GMSK, 4 TSC sets (0..3)
+	 * | . 0 1 0 X . . . | 8-PSK, 2 TSC sets (0..1) */
+	if ((mts >> 5) == 0x00) {
+		bi->mod = TRX_MOD_T_GMSK;
+		bi->tsc_set = (mts >> 3) & 0x03;
+	} else if ((mts >> 4) == 0x02) {
+		bi->mod = TRX_MOD_T_8PSK;
+		bi->tsc_set = (mts >> 3) & 0x01;
 	} else {
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
-			"Indicated modulation 0x%02x is not supported\n", mts & 0b1110);
+		LOGPPHI(phy_inst, DTRX, LOGL_ERROR,
+			"Rx TRXD PDU with unknown or not supported "
+			"modulation (MTS=0x%02x)\n", mts);
 		return -ENOTSUP;
 	}
 
 	/* Training Sequence Code */
-	bi->tsc = buf[0] & 0b111;
-	bi->flags |= TRX_BI_F_TS_INFO;
+	bi->tsc = mts & 0x07;
+
+	bi->flags |= (TRX_BI_F_MOD_TYPE | TRX_BI_F_TS_INFO);
+
+	return 0;
+}
+
+/* TRXD header dissector for version 0x01 */
+static int trx_data_handle_hdr_v1(struct phy_instance *phy_inst,
+				  struct trx_ul_burst_ind *bi,
+				  const uint8_t *buf, size_t buf_len)
+{
+	int rc;
+
+	/* Parse TRXDv0 specific header part */
+	trx_data_handle_hdr_v0_part(bi, buf);
+	buf += TRX_UL_V0HDR_LEN;
+
+	/* MTS (Modulation and Training Sequence) */
+	rc = trx_data_parse_mts(phy_inst, bi, buf[0]);
+	if (rc < 0)
+		return rc;
 
-skip_mts:
 	/* C/I: Carrier-to-Interference ratio (in centiBels) */
 	bi->ci_cb = (int16_t) osmo_load16be(buf + 1);
 	bi->flags |= TRX_BI_F_CI_CB;
@@ -808,31 +872,73 @@ skip_mts:
 	return TRX_UL_V1HDR_LEN;
 }
 
-/* TRXD burst handler for header version 0 */
-static int trx_data_handle_burst_v0(struct trx_l1h *l1h,
-				    struct trx_ul_burst_ind *bi,
-				    const uint8_t *buf, size_t buf_len)
+/* TRXD header dissector for version 0x01 */
+static int trx_data_handle_pdu_v2(struct phy_instance *phy_inst,
+				  struct trx_ul_burst_ind *bi,
+				  const uint8_t *buf, size_t buf_len)
+{
+	int rc;
+
+	/* TDMA timeslot number (other bits are RFU) */
+	bi->tn = buf[0] & 0x07;
+
+	if (buf[1] & (1 << 7)) /* BATCH.ind */
+		bi->flags |= TRX_BI_F_BATCH_IND;
+	if (buf[1] & (1 << 6)) /* VAMOS.ind */
+		bi->flags |= TRX_BI_F_SHADOW_IND;
+
+	/* TRX (RF channel) number */
+	bi->trx_num = buf[1] & 0x3f;
+	bi->flags |= TRX_BI_F_TRX_NUM;
+
+	/* MTS (Modulation and Training Sequence) */
+	rc = trx_data_parse_mts(phy_inst, bi, buf[2]);
+	if (rc < 0)
+		return rc;
+
+	bi->rssi = -(int8_t)buf[3];
+	bi->toa256 = (int16_t) osmo_load16be(&buf[4]);
+	bi->ci_cb = (int16_t) osmo_load16be(&buf[6]);
+	bi->flags |= TRX_BI_F_CI_CB;
+
+	/* TDMA frame number is absent in batched PDUs */
+	if (bi->_num_pdus == 0) {
+		if (buf_len < sizeof(bi->fn) + TRX_UL_V2HDR_LEN) {
+			LOGPPHI(phy_inst, DTRX, LOGL_ERROR,
+				"Rx malformed TRXDv2 PDU: not enough bytes "
+				"to parse TDMA frame number\n");
+			return -EINVAL;
+		}
+
+		bi->fn = osmo_load32be(buf + TRX_UL_V2HDR_LEN);
+		return TRX_UL_V2HDR_LEN + sizeof(bi->fn);
+	}
+
+	return TRX_UL_V2HDR_LEN;
+}
+
+/* TRXD burst handler (version independent) */
+static int trx_data_handle_burst(struct trx_ul_burst_ind *bi,
+				 const uint8_t *buf, size_t buf_len)
 {
 	size_t i;
 
-	/* Verify burst length */
-	switch (buf_len) {
-	/* Legacy transceivers append two padding bytes */
-	case EGPRS_BURST_LEN + 2:
-	case GSM_BURST_LEN + 2:
-		bi->burst_len = buf_len - 2;
-		break;
-	case EGPRS_BURST_LEN:
-	case GSM_BURST_LEN:
-		bi->burst_len = buf_len;
-		break;
-
-	default:
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_NOTICE,
-			"Rx TRXD message with odd burst length %zu\n", buf_len);
-		return -EINVAL;
+	/* NOPE.ind contains no burst */
+	if (bi->flags & TRX_BI_F_NOPE_IND) {
+		bi->burst_len = 0;
+		return 0;
 	}
 
+	/* Modulation types defined in 3GPP TS 45.002 */
+	static const size_t bl[] = {
+		[TRX_MOD_T_GMSK] = 148, /* 1 bit per symbol */
+		[TRX_MOD_T_8PSK] = 444, /* 3 bits per symbol */
+	};
+
+	bi->burst_len = bl[bi->mod];
+	if (buf_len < bi->burst_len)
+		return -EINVAL;
+
 	/* Convert unsigned soft-bits [254..0] to soft-bits [-127..127] */
 	for (i = 0; i < bi->burst_len; i++) {
 		if (buf[i] == 255)
@@ -844,30 +950,6 @@ static int trx_data_handle_burst_v0(struct trx_l1h *l1h,
 	return 0;
 }
 
-/* TRXD burst handler for header version 1 */
-static int trx_data_handle_burst_v1(struct trx_l1h *l1h,
-				    struct trx_ul_burst_ind *bi,
-				    const uint8_t *buf, size_t buf_len)
-{
-	/* Modulation types defined in 3GPP TS 45.002 */
-	static const size_t bl[] = {
-		[TRX_BURST_GMSK] = 148, /* 1 bit per symbol */
-		[TRX_BURST_8PSK] = 444, /* 3 bits per symbol */
-	};
-
-	/* Verify burst length */
-	if (bl[bi->bt] != buf_len) {
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_NOTICE,
-			"Rx TRXD message with odd burst length %zu, "
-			"expected %zu\n", buf_len, bl[bi->bt]);
-		return -EINVAL;
-	}
-
-	/* The burst format is the same as for version 0.
-	 * NOTE: other modulation types to be handled separately. */
-	return trx_data_handle_burst_v0(l1h, bi, buf, buf_len);
-}
-
 static const char *trx_data_desc_msg(const struct trx_ul_burst_ind *bi)
 {
 	struct osmo_strbuf sb;
@@ -875,8 +957,8 @@ static const char *trx_data_desc_msg(const struct trx_ul_burst_ind *bi)
 
 	/* Modulation types defined in 3GPP TS 45.002 */
 	static const char *mod_names[] = {
-		[TRX_BURST_GMSK] = "GMSK",
-		[TRX_BURST_8PSK] = "8-PSK",
+		[TRX_MOD_T_GMSK] = "GMSK",
+		[TRX_MOD_T_8PSK] = "8-PSK",
 	};
 
 	/* Initialize the string buffer */
@@ -885,6 +967,10 @@ static const char *trx_data_desc_msg(const struct trx_ul_burst_ind *bi)
 	/* Common TDMA parameters */
 	OSMO_STRBUF_PRINTF(sb, "tn=%u fn=%u", bi->tn, bi->fn);
 
+	/* TRX (RF channel number) */
+	if (bi->flags & TRX_BI_F_TRX_NUM)
+		OSMO_STRBUF_PRINTF(sb, " trx_num=%u", bi->trx_num);
+
 	/* RSSI and ToA256 */
 	OSMO_STRBUF_PRINTF(sb, " rssi=%d toa256=%d", bi->rssi, bi->toa256);
 
@@ -898,7 +984,7 @@ static const char *trx_data_desc_msg(const struct trx_ul_burst_ind *bi)
 
 	/* Modulation and TSC set */
 	if (bi->flags & TRX_BI_F_MOD_TYPE)
-		OSMO_STRBUF_PRINTF(sb, " mod=%s", mod_names[bi->bt]);
+		OSMO_STRBUF_PRINTF(sb, " mod=%s", mod_names[bi->mod]);
 
 	/* Training Sequence Code */
 	if (bi->flags & TRX_BI_F_TS_INFO)
@@ -910,197 +996,110 @@ static const char *trx_data_desc_msg(const struct trx_ul_burst_ind *bi)
 	return buf;
 }
 
-/* Parse TRXD message from transceiver, compose an UL burst indication.
- *
- * This message contains a demodulated Uplink burst with fixed-size
- * header preceding the burst bits. The header consists of the common
- * and message specific part.
- *
- *   +---------------+-----------------+------------+
- *   | common header | specific header | burst bits |
- *   +---------------+-----------------+------------+
- *
- * Common header is the same as for Downlink message:
- *
- *   +-----------------+----------------+-------------------+
- *   | VER (1/2 octet) | TN (1/2 octet) | FN (4 octets, BE) |
- *   +-----------------+----------------+-------------------+
- *
- * and among with TDMA parameters, contains the version indicator:
- *
- *   +-----------------+------------------------+
- *   | 7 6 5 4 3 2 1 0 | bit numbers            |
- *   +-----------------+------------------------+
- *   | X X X X . . . . | header version (0..15) |
- *   +-----------------+------------------------+
- *   | . . . . . X X X | TDMA TN (0..7)         |
- *   +-----------------+------------------------+
- *   | . . . . X . . . | RESERVED (0)           |
- *   +-----------------+------------------------+
- *
- * which is encoded in 4 MSB bits of the first octet, which used to be
- * zero-initialized due to the value range of TDMA TN. Therefore, the
- * old header format has implicit version 0x00.
- *
- * The message specific header has the following structure:
- *
- * == Version 0x00
- *
- *   +------+-----+--------------------+
- *   | RSSI | ToA | soft-bits (254..0) |
- *   +------+-----+--------------------+
- *
- * == Version 0x01
- *
- *   +------+-----+-----+-----+--------------------+
- *   | RSSI | ToA | MTS | C/I | soft-bits (254..0) |
- *   +------+-----+-----+-----+--------------------+
- *
- * where:
- *
- *   - RSSI (1 octet) - Received Signal Strength Indication
- *     encoded without the negative sign.
- *   - ToA (2 octets) - Timing of Arrival in units of 1/256
- *     of symbol (big endian).
- *   - MTS (1 octet)  - Modulation and Training Sequence info.
- *   - C/I (2 octets) - Carrier-to-Interference ratio (big endian).
- *
- * == Coding of MTS: Modulation and Training Sequence info
- *
- * 3GPP TS 45.002 version 15.1.0 defines several modulation types,
- * and a few sets of training sequences for each type. The most
- * common are GMSK and 8-PSK (which is used in EDGE).
- *
- *   +-----------------+---------------------------------------+
- *   | 7 6 5 4 3 2 1 0 | bit numbers (value range)             |
- *   +-----------------+---------------------------------------+
- *   | . . . . . X X X | Training Sequence Code (0..7)         |
- *   +-----------------+---------------------------------------+
- *   | . X X X X . . . | Modulation, TS set number (see below) |
- *   +-----------------+---------------------------------------+
- *   | X . . . . . . . | IDLE / nope frame indication (0 or 1) |
- *   +-----------------+---------------------------------------+
- *
- * The bit number 7 (MSB) is set to high when either nothing has been
- * detected, or during IDLE frames, so we can deliver noise levels,
- * and avoid clock gaps on the L1 side. Other bits are ignored,
- * and should be set to low (0) in this case. L16 shall be set to 0x00.
- *
- * == Coding of modulation and TS set number
- *
- * GMSK has 4 sets of training sequences (see tables 5.2.3a-d),
- * while 8-PSK (see tables 5.2.3f-g) and the others have 2 sets.
- * Access and Synchronization bursts also have several synch.
- * sequences.
- *
- *   +-----------------+---------------------------------------+
- *   | 7 6 5 4 3 2 1 0 | bit numbers (value range)             |
- *   +-----------------+---------------------------------------+
- *   | . 0 0 X X . . . | GMSK, 4 TS sets (0..3)                |
- *   +-----------------+---------------------------------------+
- *   | . 0 1 0 X . . . | 8-PSK, 2 TS sets (0..1)               |
- *   +-----------------+---------------------------------------+
- *   | . 0 1 1 X . . . | AQPSK, 2 TS sets (0..1)               |
- *   +-----------------+---------------------------------------+
- *   | . 1 0 0 X . . . | 16QAM, 2 TS sets (0..1)               |
- *   +-----------------+---------------------------------------+
- *   | . 1 0 1 X . . . | 32QAM, 2 TS sets (0..1)               |
- *   +-----------------+---------------------------------------+
- *   | . 1 1 1 X . . . | RESERVED (0)                          |
- *   +-----------------+---------------------------------------+
- *
- * NOTE: we only support GMSK and 8-PSK.
- *
- * == C/I: Carrier-to-Interference ratio
- *
- * The C/I value can be computed from the training sequence of each
- * burst, where we can compare the "ideal" training sequence with
- * the actual training sequence and then express that in centiBels.
- *
- * == Coding of the burst bits
- *
- * Unlike to be transmitted bursts, the received bursts are designated
- * using the soft-bits notation, so the receiver can indicate its
- * assurance from 0 to -127 that a given bit is 1, and from 0 to +127
- * that a given bit is 0.
- *
- * Each soft-bit (-127..127) of the burst is encoded as an unsigned
- * value in range (254..0) respectively using the constant shift.
- *
- */
+/* TRXD buffer used by Rx/Tx handlers */
+static uint8_t trx_data_buf[TRXD_MSG_BUF_SIZE];
+
+/* Parse TRXD message from transceiver, compose an UL burst indication. */
 static int trx_data_read_cb(struct osmo_fd *ofd, unsigned int what)
 {
+	const uint8_t *buf = &trx_data_buf[0];
 	struct trx_l1h *l1h = ofd->data;
-	uint8_t buf[TRX_DATA_MSG_MAX_LEN];
 	struct trx_ul_burst_ind bi;
 	ssize_t hdr_len, buf_len;
-	uint8_t hdr_ver;
-	int rc;
+	uint8_t pdu_ver;
 
-	buf_len = recv(ofd->fd, buf, sizeof(buf), 0);
+	buf_len = recv(ofd->fd, trx_data_buf, sizeof(trx_data_buf), 0);
 	if (buf_len <= 0) {
-		strerror_r(errno, (char *)buf, sizeof(buf));
+		strerror_r(errno, (char *) trx_data_buf, sizeof(trx_data_buf));
 		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
-			"recv() failed on TRXD with rc=%zd (%s)\n", buf_len, buf);
+			"recv() failed on TRXD with rc=%zd (%s)\n",
+			buf_len, trx_data_buf);
 		return buf_len;
 	}
 
-	/* Pre-clean (initialize) the flags */
-	bi.flags = 0x00;
+	/* Parse PDU version first */
+	pdu_ver = buf[0] >> 4;
 
-	/* Parse the header depending on its version */
-	hdr_ver = buf[0] >> 4;
-	switch (hdr_ver) {
-	case 0:
-		/* Legacy protocol has no version indicator */
-		hdr_len = trx_data_handle_hdr_v0(l1h, &bi, buf, buf_len);
-		break;
-	case 1:
-		hdr_len = trx_data_handle_hdr_v1(l1h, &bi, buf, buf_len);
-		break;
-	default:
+	/* Make sure that PDU version matches our expectations */
+	if (pdu_ver != l1h->config.trxd_pdu_ver_use) {
 		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
-			"TRXD header version %u is not supported\n", hdr_ver);
-		return -ENOTSUP;
+			"Rx TRXD PDU with unexpected version %u (expected %u)\n",
+			pdu_ver, l1h->config.trxd_pdu_ver_use);
+		return -EIO;
 	}
 
-	/* Header parsing error */
-	if (hdr_len < 0)
-		return hdr_len;
+	/* We're about to parse the first PDU */
+	bi._num_pdus = 0;
 
-	if (bi.flags & TRX_BI_F_NOPE_IND) {
-		bi.burst_len = 0;
-		goto skip_burst;
-	}
+	/* Starting from TRXDv2, there can be batched PDUs */
+	do {
+		/* (Re)initialize the flags */
+		bi.flags = 0x00;
 
-	/* We're done with the header now */
-	buf_len -= hdr_len;
+		/* Make sure that we have enough bytes to parse the header */
+		if (buf_len < trx_data_rx_hdr_len[pdu_ver]) {
+			LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
+				"Rx malformed TRXDv%u PDU: len=%zd < expected %u\n",
+				pdu_ver, buf_len, trx_data_rx_hdr_len[pdu_ver]);
+			return -EINVAL;
+		}
 
-	/* Handle burst bits */
-	switch (hdr_ver) {
-	case 0:
-		rc = trx_data_handle_burst_v0(l1h, &bi, buf + hdr_len, buf_len);
-		break;
-	case 1:
-		rc = trx_data_handle_burst_v1(l1h, &bi, buf + hdr_len, buf_len);
-		break;
-	default:
-		/* Shall not happen, just to make GCC happy */
-		OSMO_ASSERT(0);
-	}
+		/* Parse header depending on the PDU version */
+		switch (pdu_ver) {
+		case 0: /* TRXDv0 */
+			hdr_len = trx_data_handle_hdr_v0(l1h->phy_inst, &bi, buf, buf_len);
+			break;
+		case 1: /* TRXDv1 */
+			hdr_len = trx_data_handle_hdr_v1(l1h->phy_inst, &bi, buf, buf_len);
+			break;
+		case 2: /* TRXDv2 */
+			hdr_len = trx_data_handle_pdu_v2(l1h->phy_inst, &bi, buf, buf_len);
+			break;
+		default:
+			/* Shall not happen */
+			OSMO_ASSERT(0);
+		}
 
-	/* Burst parsing error */
-	if (rc < 0)
-		return rc;
+		/* Header parsing error */
+		if (hdr_len < 0)
+			return hdr_len;
 
-skip_burst:
-	/* Print header & burst info */
-	LOGPPHI(l1h->phy_inst, DTRX, LOGL_DEBUG, "Rx %s (hdr_ver=%u): %s\n",
-		(bi.flags & TRX_BI_F_NOPE_IND) ? "NOPE.ind" : "UL burst",
-		hdr_ver, trx_data_desc_msg(&bi));
+		if (bi.fn >= GSM_TDMA_HYPERFRAME) {
+			LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
+				"Rx malformed TRXDv%u PDU: illegal TDMA fn=%u\n",
+				pdu_ver, bi.fn);
+			return -EINVAL;
+		}
 
-	/* feed received burst into scheduler code */
-	trx_sched_route_burst_ind(&bi, &l1h->l1s);
+		/* We're done with the header now */
+		buf_len -= hdr_len;
+		buf += hdr_len;
+
+		/* Calculate burst length and parse it (if present) */
+		if (trx_data_handle_burst(&bi, buf, buf_len) != 0) {
+			LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
+				"Rx malformed TRXDv%u PDU: odd burst length=%zd\n",
+				pdu_ver, buf_len);
+			return -EINVAL;
+		}
+
+		/* We're done with the burst bits now */
+		buf_len -= bi.burst_len;
+		buf += bi.burst_len;
+
+		/* Print header & burst info */
+		LOGPPHI(l1h->phy_inst, DTRX, LOGL_DEBUG, "Rx %s (pdu_ver=%u): %s\n",
+			(bi.flags & TRX_BI_F_NOPE_IND) ? "NOPE.ind" : "UL burst",
+			pdu_ver, trx_data_desc_msg(&bi));
+
+		/* Number of processed PDUs */
+		bi._num_pdus++;
+
+		/* feed received burst into scheduler code */
+		TRACE(OSMO_BTS_TRX_UL_DATA_START(l1h->phy_inst->trx->nr, bi.tn, bi.fn));
+		trx_sched_route_burst_ind(l1h->phy_inst->trx, &bi);
+		TRACE(OSMO_BTS_TRX_UL_DATA_DONE(l1h->phy_inst->trx->nr, bi.tn, bi.fn));
+	} while (bi.flags & TRX_BI_F_BATCH_IND);
 
 	return 0;
 }
@@ -1111,50 +1110,93 @@ skip_burst:
  *  \returns 0 on success; negative on error */
 int trx_if_send_burst(struct trx_l1h *l1h, const struct trx_dl_burst_req *br)
 {
-	ssize_t snd_len;
-	uint8_t hdr_ver = l1h->config.trxd_hdr_ver_use;
-	uint8_t buf[TRX_DATA_MSG_MAX_LEN];
+	uint8_t pdu_ver = l1h->config.trxd_pdu_ver_use;
+	static uint8_t *buf = &trx_data_buf[0];
+	static uint8_t *last_pdu = NULL;
+	static unsigned int pdu_num = 0;
+	ssize_t snd_len, buf_len;
 
-	if ((br->burst_len != GSM_BURST_LEN) && (br->burst_len != EGPRS_BURST_LEN)) {
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR, "Tx burst length %zu invalid\n",
-			br->burst_len);
-		return -1;
-	}
-
-	LOGPPHI(l1h->phy_inst, DTRX, LOGL_DEBUG,
-		"Tx burst (hdr_ver=%u): tn=%u fn=%u att=%u\n",
-		hdr_ver, br->tn, br->fn, br->att);
-
-	switch (hdr_ver) {
-	case 0:
-	case 1:
-		/* Both versions have the same header format */
-		break;
-
-	default:
+	/* Make sure that the PHY is powered on */
+	if (!trx_if_powered(l1h)) {
 		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
-			"Requested TRXD header version %u is not supported\n", hdr_ver);
-		return -ENOTSUP;
+			"Ignoring Tx data, transceiver is powered off\n");
+		return -ENODEV;
 	}
 
-	buf[0] = ((hdr_ver & 0x0f) << 4) | br->tn;
-	osmo_store32be(br->fn, buf + 1);
-	buf[5] = br->att;
+	/* Burst batching breaker */
+	if (br == NULL) {
+		if (pdu_num > 0)
+			goto sendall;
+		return -ENOMSG;
+	}
+
+	/* Pointer to the last encoded PDU */
+	last_pdu = &buf[0];
+
+	switch (pdu_ver) {
+	/* Both versions have the same PDU format */
+	case 0: /* TRXDv0 */
+	case 1: /* TRXDv1 */
+		buf[0] = ((pdu_ver & 0x0f) << 4) | br->tn;
+		osmo_store32be(br->fn, buf + 1);
+		buf[5] = br->att;
+		buf += 6;
+		break;
+	case 2: /* TRXDv2 */
+		buf[0] = br->tn;
+		/* BATCH.ind will be unset in the last PDU */
+		buf[1] = (br->trx_num & 0x3f) | (1 << 7);
+		buf[2] = trx_data_mod_val[br->mod]
+		       | (br->tsc_set << 3)
+		       | (br->tsc & 0x07);
+		buf[3] = br->att;
+		buf[4] = (uint8_t) br->scpir;
+		buf[5] = buf[6] = buf[7] = 0x00; /* Spare */
+		/* Some fields are not present in batched PDUs */
+		if (pdu_num == 0) {
+			buf[0] |= (pdu_ver & 0x0f) << 4;
+			osmo_store32be(br->fn, buf + 8);
+			buf += 4;
+		}
+		buf += 8;
+		break;
+	default:
+		/* Shall not happen */
+		OSMO_ASSERT(0);
+	}
 
 	/* copy ubits {0,1} */
-	memcpy(buf + 6, br->burst, br->burst_len);
+	memcpy(buf, br->burst, br->burst_len);
+	buf += br->burst_len;
 
-	/* we must be sure that TRX is on */
-	if (trx_if_powered(l1h)) {
-		snd_len = send(l1h->trx_ofd_data.fd, buf, br->burst_len + 6, 0);
-		if (snd_len <= 0) {
-			strerror_r(errno, (char *)buf, sizeof(buf));
-			LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
-				"send() failed on TRXD with rc=%zd (%s)\n", snd_len, buf);
-			return -2;
-		}
-	} else
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR, "Ignoring TX data, transceiver powered off.\n");
+	/* One more PDU in the buffer */
+	pdu_num++;
+
+	/* TRXDv2: wait for the batching breaker */
+	if (pdu_ver >= 2)
+		return 0;
+
+sendall:
+	LOGPPHI(l1h->phy_inst, DTRX, LOGL_DEBUG,
+		"Tx TRXDv%u datagram with %u PDU(s)\n",
+		pdu_ver, pdu_num);
+
+	/* TRXDv2: unset BATCH.ind in the last PDU */
+	if (pdu_ver >= 2)
+		last_pdu[1] &= ~(1 << 7);
+
+	buf_len = buf - &trx_data_buf[0];
+	buf = &trx_data_buf[0];
+	pdu_num = 0;
+
+	snd_len = send(l1h->trx_ofd_data.fd, trx_data_buf, buf_len, 0);
+	if (snd_len <= 0) {
+		strerror_r(errno, (char *) trx_data_buf, sizeof(trx_data_buf));
+		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
+			"send() failed on TRXD with rc=%zd (%s)\n",
+			snd_len, trx_data_buf);
+		return -2;
+	}
 
 	return 0;
 }
@@ -1177,12 +1219,17 @@ void trx_if_flush(struct trx_l1h *l1h)
 		talloc_free(tcm);
 	}
 	talloc_free(l1h->last_acked);
+	l1h->last_acked = NULL;
+
+	/* Tx queue is now empty, so there's no point in keeping the retrans timer armed: */
+	if (osmo_timer_pending(&l1h->trx_ctrl_timer))
+		osmo_timer_del(&l1h->trx_ctrl_timer);
 }
 
 /*! close the TRX for given handle (data + control socket) */
 void trx_if_close(struct trx_l1h *l1h)
 {
-	LOGPPHI(l1h->phy_inst, DTRX, LOGL_NOTICE, "Close transceiver\n");
+	LOGPPHI(l1h->phy_inst, DTRX, LOGL_NOTICE, "Closing TRXC/TRXD connections\n");
 
 	trx_if_flush(l1h);
 
@@ -1213,10 +1260,7 @@ static int trx_if_open(struct trx_l1h *l1h)
 	struct phy_link *plink = pinst->phy_link;
 	int rc;
 
-	LOGPPHI(l1h->phy_inst, DTRX, LOGL_NOTICE, "Open transceiver\n");
-
-	/* initialize ctrl queue */
-	INIT_LLIST_HEAD(&l1h->trx_ctrl_list);
+	LOGPPHI(pinst, DTRX, LOGL_NOTICE, "Opening TRXC/TRXD connections\n");
 
 	/* open sockets */
 	rc = trx_udp_open(l1h, &l1h->trx_ofd_ctrl,
@@ -1225,20 +1269,16 @@ static int trx_if_open(struct trx_l1h *l1h)
 			  plink->u.osmotrx.remote_ip,
 			  compute_port(pinst, 1, 0), trx_ctrl_read_cb);
 	if (rc < 0)
-		goto err;
+		return rc;
 	rc = trx_udp_open(l1h, &l1h->trx_ofd_data,
 			  plink->u.osmotrx.local_ip,
 			  compute_port(pinst, 0, 1),
 			  plink->u.osmotrx.remote_ip,
 			  compute_port(pinst, 1, 1), trx_data_read_cb);
 	if (rc < 0)
-		goto err;
+		return rc;
 
 	return 0;
-
-err:
-	trx_if_close(l1h);
-	return rc;
 }
 
 /*! close the control + burst data sockets for one phy_instance */
@@ -1247,7 +1287,8 @@ static void trx_phy_inst_close(struct phy_instance *pinst)
 	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
 
 	trx_if_close(l1h);
-	trx_sched_exit(&l1h->l1s);
+	if (pinst->trx)
+		trx_sched_clean(pinst->trx);
 }
 
 /*! open the control + burst data sockets for one phy_instance */
@@ -1260,11 +1301,9 @@ static int trx_phy_inst_open(struct phy_instance *pinst)
 	if (!l1h)
 		return -EINVAL;
 
-	rc = trx_sched_init(&l1h->l1s, pinst->trx);
-	if (rc < 0) {
-		LOGPPHI(l1h->phy_inst, DL1C, LOGL_FATAL, "Cannot initialize scheduler\n");
-		return -EIO;
-	}
+	/* PHY instance may be not associated with a TRX instance */
+	if (pinst->trx != NULL)
+		trx_sched_init(pinst->trx);
 
 	rc = trx_if_open(l1h);
 	if (rc < 0) {
@@ -1318,6 +1357,23 @@ cleanup:
 	return -1;
 }
 
+/*! close the PHY link using TRX protocol */
+int bts_model_phy_link_close(struct phy_link *plink)
+{
+	bool clock_stopped = false;
+	struct phy_instance *pinst;
+	llist_for_each_entry(pinst, &plink->instances, list) {
+		if (!clock_stopped) {
+			clock_stopped = true;
+			trx_sched_clock_stopped(pinst->trx->bts);
+		}
+		trx_phy_inst_close(pinst);
+	}
+	trx_udp_close(&plink->u.osmotrx.trx_ofd_clk);
+	phy_link_state_set(plink, PHY_LINK_SHUTDOWN);
+	return 0;
+}
+
 /*! determine if the TRX for given handle is powered up */
 int trx_if_powered(struct trx_l1h *l1h)
 {
