@@ -1,4 +1,4 @@
-/* VTY interface for sysmoBTS */
+/* VTY interface for osmo-bts-trx */
 
 /* (C) 2013 by Andreas Eversberg <jolly@eversberg.eu>
  *
@@ -57,6 +57,7 @@ DEFUN(show_transceiver, show_transceiver_cmd, "show transceiver",
 {
 	struct gsm_bts_trx *trx;
 	struct trx_l1h *l1h;
+	unsigned int tn;
 
 	llist_for_each_entry(trx, &g_bts->trx_list, list) {
 		struct phy_instance *pinst = trx_phy_instance(trx);
@@ -86,7 +87,23 @@ DEFUN(show_transceiver, show_transceiver_cmd, "show transceiver",
 			vty_out(vty, " bsic   : %d%s", l1h->config.bsic,
 				VTY_NEWLINE);
 		else
-			vty_out(vty, " bisc   : undefined%s", VTY_NEWLINE);
+			vty_out(vty, " bsic   : undefined%s", VTY_NEWLINE);
+
+		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+			const struct gsm_bts_trx_ts *ts = &trx->ts[tn];
+			const struct l1sched_ts *l1ts = ts->priv;
+			const struct trx_sched_multiframe *mf;
+
+			mf = &trx_sched_multiframes[l1ts->mf_index];
+
+			vty_out(vty, "  timeslot #%u (%s)%s",
+				tn, mf->name, VTY_NEWLINE);
+			vty_out(vty, "    pending DL prims    : %u%s",
+				llist_count(&l1ts->dl_prims), VTY_NEWLINE);
+			vty_out(vty, "    interference        : %ddBm%s",
+				l1ts->chan_state[TRXC_IDLE].meas.interf_avg,
+				VTY_NEWLINE);
+		}
 	}
 
 	return CMD_SUCCESS;
@@ -99,17 +116,23 @@ static void show_phy_inst_single(struct vty *vty, struct phy_instance *pinst)
 	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
 	struct gsm_bts_trx *trx = pinst->trx;
 
-	vty_out(vty, "PHY Instance %s%s",
-		phy_instance_name(pinst), VTY_NEWLINE);
+	vty_out(vty, "PHY Instance '%s': bound to %s%s",
+		phy_instance_name(pinst),
+		gsm_trx_name(trx),
+		VTY_NEWLINE);
+
+	if (trx != NULL) {
+		const int actual = get_p_actual_mdBm(trx, trx->power_params.p_total_tgt_mdBm);
+		const int max = get_p_max_out_mdBm(trx);
+		vty_out(vty, " tx-attenuation : %d dB%s",
+			(max - actual) / 1000, VTY_NEWLINE);
+	}
 
 	if (l1h->config.rxgain_valid)
 		vty_out(vty, " rx-gain        : %d dB%s",
 			l1h->config.rxgain, VTY_NEWLINE);
 	else
 		vty_out(vty, " rx-gain        : undefined%s", VTY_NEWLINE);
-	vty_out(vty, " tx-attenuation : %d dB%s",
-		(get_p_max_out_mdBm(trx) - get_p_actual_mdBm(trx, trx->power_params.p_total_tgt_mdBm))/1000,
-		VTY_NEWLINE);
 	if (l1h->config.maxdly_valid)
 		vty_out(vty, " maxdly : %d%s", l1h->config.maxdly,
 			VTY_NEWLINE);
@@ -121,16 +144,23 @@ static void show_phy_inst_single(struct vty *vty, struct phy_instance *pinst)
 	else
 		vty_out(vty, " maxdlynb : undefined%s", VTY_NEWLINE);
 	for (tn = 0; tn < TRX_NR_TS; tn++) {
-		if (!((1 << tn) & l1h->config.slotmask))
+		if (!((1 << tn) & l1h->config.slotmask)) {
 			vty_out(vty, " slot #%d: unsupported%s", tn,
 				VTY_NEWLINE);
-		else if (l1h->config.slottype_valid[tn])
-			vty_out(vty, " slot #%d: type %d%s", tn,
-				l1h->config.slottype[tn],
-				VTY_NEWLINE);
-		else
+			continue;
+		} else if (!l1h->config.setslot_valid[tn]) {
 			vty_out(vty, " slot #%d: undefined%s", tn,
 				VTY_NEWLINE);
+			continue;
+		}
+
+		vty_out(vty, " slot #%d: type %d", tn,
+			l1h->config.setslot[tn].slottype);
+		if (l1h->config.setslot[tn].tsc_valid)
+			vty_out(vty, " TSC-s%dc%d",
+				l1h->config.setslot[tn].tsc_set,
+				l1h->config.setslot[tn].tsc_val);
+		vty_out(vty, "%s", VTY_NEWLINE);
 	}
 }
 
@@ -235,18 +265,12 @@ DEFUN_DEPRECATED(cfg_phy_no_timing_advance_loop, cfg_phy_no_timing_advance_loop_
 	return CMD_SUCCESS;
 }
 
-DEFUN_ATTR(cfg_phyinst_maxdly, cfg_phyinst_maxdly_cmd,
-	   "osmotrx maxdly <0-31>",
-	   OSMOTRX_STR
-	   "Set the maximum acceptable delay of an Access Burst (in GSM symbols)."
-	   " Access Burst is the first burst a mobile transmits in order to establish"
-	   " a connection and it is used to estimate Timing Advance (TA) which is"
-	   " then applied to Normal Bursts to compensate for signal delay due to"
-	   " distance. So changing this setting effectively changes maximum range of"
-	   " the cell, because if we receive an Access Burst with a delay higher than"
-	   " this value, it will be ignored and connection is dropped.\n"
-	   "GSM symbols (approx. 1.1km per symbol)\n",
-	   CMD_ATTR_IMMEDIATE)
+DEFUN_USRATTR(cfg_phyinst_maxdly, cfg_phyinst_maxdly_cmd,
+	      X(BTS_VTY_TRX_POWERCYCLE),
+	      "osmotrx maxdly <0-63>",
+	      OSMOTRX_STR
+	      "Set the maximum acceptable delay of an Access Burst\n"
+	      "Delay in GSMK symbol periods (approx. 550m per symbol)\n")
 {
 	struct phy_instance *pinst = vty->index;
 	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
@@ -254,25 +278,17 @@ DEFUN_ATTR(cfg_phyinst_maxdly, cfg_phyinst_maxdly_cmd,
 	l1h->config.maxdly = atoi(argv[0]);
 	l1h->config.maxdly_valid = 1;
 	l1h->config.maxdly_sent = false;
-	l1if_provision_transceiver_trx(l1h);
 
 	return CMD_SUCCESS;
 }
 
-DEFUN_ATTR(cfg_phyinst_maxdlynb, cfg_phyinst_maxdlynb_cmd,
-	   "osmotrx maxdlynb <0-31>",
-	   OSMOTRX_STR
-	   "Set the maximum acceptable delay of a Normal Burst (in GSM symbols)."
-	   " USE FOR TESTING ONLY, DON'T CHANGE IN PRODUCTION USE!"
-	   " During normal operation, Normal Bursts delay are controlled by a Timing"
-	   " Advance control loop and thus Normal Bursts arrive to a BTS with no more"
-	   " than a couple GSM symbols, which is already taken into account in osmo-trx."
-	   " So changing this setting will have no effect in production installations"
-	   " except increasing osmo-trx CPU load. This setting is only useful when"
-	   " testing with a transmitter which can't precisely synchronize to the BTS"
-	   " downlink signal, like e.g. R&S CMD57.\n"
-	   "GSM symbols (approx. 1.1km per symbol)\n",
-	   CMD_ATTR_IMMEDIATE)
+DEFUN_ATTR_USRATTR(cfg_phyinst_maxdlynb, cfg_phyinst_maxdlynb_cmd,
+		   CMD_ATTR_HIDDEN, /* expert mode command */
+		   X(BTS_VTY_TRX_POWERCYCLE),
+		   "osmotrx maxdlynb <0-63>",
+		   OSMOTRX_STR
+		   "Set the maximum acceptable delay of a Normal Burst\n"
+		   "Delay in GMSK symbol periods (approx. 550m per symbol)\n")
 {
 	struct phy_instance *pinst = vty->index;
 	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
@@ -280,7 +296,6 @@ DEFUN_ATTR(cfg_phyinst_maxdlynb, cfg_phyinst_maxdlynb_cmd,
 	l1h->config.maxdlynb = atoi(argv[0]);
 	l1h->config.maxdlynb_valid = 1;
 	l1h->config.maxdlynb_sent = false;
-	l1if_provision_transceiver_trx(l1h);
 
 	return CMD_SUCCESS;
 }
@@ -357,12 +372,12 @@ DEFUN_ATTR(cfg_phy_rts_advance, cfg_phy_rts_advance_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN_ATTR(cfg_phyinst_rxgain, cfg_phyinst_rxgain_cmd,
-	   "osmotrx rx-gain <0-50>",
-	   OSMOTRX_STR
-	   "Set the receiver gain in dB\n"
-	   "Gain in dB\n",
-	   CMD_ATTR_IMMEDIATE)
+DEFUN_USRATTR(cfg_phyinst_rxgain, cfg_phyinst_rxgain_cmd,
+	      X(BTS_VTY_TRX_POWERCYCLE),
+	      "osmotrx rx-gain <0-50>",
+	      OSMOTRX_STR
+	      "Set the receiver gain in dB\n"
+	      "Gain in dB\n")
 {
 	struct phy_instance *pinst = vty->index;
 	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
@@ -370,7 +385,6 @@ DEFUN_ATTR(cfg_phyinst_rxgain, cfg_phyinst_rxgain_cmd,
 	l1h->config.rxgain = atoi(argv[0]);
 	l1h->config.rxgain_valid = 1;
 	l1h->config.rxgain_sent = false;
-	l1if_provision_transceiver_trx(l1h);
 
 	return CMD_SUCCESS;
 }
@@ -518,15 +532,15 @@ DEFUN_USRATTR(cfg_phy_trxd_max_version, cfg_phy_trxd_max_version_cmd,
 
 	int max_ver;
 	if (strcmp(argv[0], "latest") == 0)
-		max_ver = TRX_DATA_FORMAT_VER;
+		max_ver = TRX_DATA_PDU_VER;
 	else
 		max_ver = atoi(argv[0]);
-	if (max_ver > TRX_DATA_FORMAT_VER) {
+	if (max_ver > TRX_DATA_PDU_VER) {
 		vty_out(vty, "%% Format version %d is not supported, maximum supported is %d%s",
-			max_ver, TRX_DATA_FORMAT_VER, VTY_NEWLINE);
+			max_ver, TRX_DATA_PDU_VER, VTY_NEWLINE);
 		return CMD_WARNING;
 	}
-	plink->u.osmotrx.trxd_hdr_ver_max = max_ver;
+	plink->u.osmotrx.trxd_pdu_ver_max = max_ver;
 
 	return CMD_SUCCESS;
 }
@@ -555,8 +569,8 @@ void bts_model_config_write_phy(struct vty *vty, const struct phy_link *plink)
 	if (plink->u.osmotrx.use_legacy_setbsic)
 		vty_out(vty, " osmotrx legacy-setbsic%s", VTY_NEWLINE);
 
-	if (plink->u.osmotrx.trxd_hdr_ver_max != TRX_DATA_FORMAT_VER)
-		vty_out(vty, " osmotrx trxd-max-version %d%s", plink->u.osmotrx.trxd_hdr_ver_max, VTY_NEWLINE);
+	if (plink->u.osmotrx.trxd_pdu_ver_max != TRX_DATA_PDU_VER)
+		vty_out(vty, " osmotrx trxd-max-version %d%s", plink->u.osmotrx.trxd_pdu_ver_max, VTY_NEWLINE);
 }
 
 void bts_model_config_write_phy_inst(struct vty *vty, const struct phy_instance *pinst)
